unit cMIME;

interface

uses
  cTypes,
  cInternetStandards;



{                                                                              }
{ MIME v0.01                                                                   }
{   From RFC 2045                                                              }
{                                                                              }
{ This unit is copyrighted  2000 by David Butler (david@e.co.za)              }
{                                                                              }
{ I invite you to use this unit, free of charge.                               }
{ I invite you to distibute this unit, but it must be for free.                }
{ I also invite you to contribute to its development, but do not distribute    }
{ a modified copy of this file, send modifications, suggestions and bug        }
{ reports to david@e.co.za                                                     }
{                                                                              }
{ Revision history:                                                            }
{   17/10/2000  v0.01  Moved MIME functions from unit cInternetStandards.      }
{                                                                              }



{                                                                              }
{ Decoding / Encoding functions                                                }
{                                                                              }
Function  EncodeMimeParameters (const ParamName, ParamValue : Array of String) : String;
Function  DecodeMimeParameterValue (const Parameters, ParameterName : String) : String;
Procedure SetMimeParameter (var Parameters : String; const ParamName, ParamValue : String);

Function  EncodeMimeFieldBody (const FieldBody : String;
          const ParamName, ParamValue : Array of String) : String; overload;
Function  EncodeMimeFieldBody (const FieldBody : String; const Parameters : String = '') : String; overload;
Function  DecodeMimeFieldBody (const FieldBody : String; var Parameters : String) : String;

Function  EncodeMimeHeaderField (const FieldName, FieldBody : String;
          const ParamName, ParamValue : Array of String) : String; overload;
Function  EncodeMimeHeaderField (const FieldName, FieldBody : String;
          const Parameters : String = '') : String; overload;

Procedure DecodeMimeVersionFieldBody (const FieldBody : String; var MajorVersion, MinorVersion : Integer;
          var Comment : String);
Function  EncodeMimeVersionFieldBody (const MajorVersion : Integer = 1; const MinorVersion : Integer = 1;
          const Comment : String = '') : String;

Function  EncodeMimeContentLengthField (const Length : Int64) : String;

Procedure DecodeMimeContentTypeFieldBody (const FieldBody : String; var MajorType, SubType, Parameters : String);
Function  EncodeMimeContentTypeFieldBody (const ContentType : String; const Parameters : String) : String; overload;
Function  EncodeMimeContentTypeFieldBody (const MajorType, SubType : String; const Parameters : String) : String; overload;

Function  GenerateMimeBoundary : String;



{                                                                              }
{ MIME Header fields                                                           }
{                                                                              }
type
  { AMimeHeaderField                                                           }
  AMimeHeaderField = class (AHeaderField)
    protected
    FParameters : String;

    Function  GetParameters : String; virtual;
    Procedure SetParameters (const Parameters : String); virtual;
    Function  GetParameter (const ParamName : String) : String; virtual;
    Procedure SetParameter (const ParamName : String; const ParamValue : String); virtual;

    public
    Property Parameters : String read GetParameters write SetParameters;
    Property Parameter [const ParamName : String] : String read GetParameter write SetParameter;
  end;

  { TMimeHeaderField                                                           }
  TMimeHeaderField = class (THeaderField)
    protected
    FParameters : String;

    Function  GetParameters : String; virtual;
    Procedure SetParameters (const Parameters : String); virtual;
    Function  GetParameter (const ParamName : String) : String; virtual;
    Procedure SetParameter (const ParamName : String; const ParamValue : String); virtual;

    public
    Property Parameters : String read GetParameters write SetParameters;
    Property Parameter [const ParamName : String] : String read GetParameter write SetParameter;
  end;

  { TMimeVersionField                                                          }
  TMimeVersionField = class (AMimeHeaderField)
    protected
    Function  GetName : String; override;
    Procedure SetBody (const FieldBody : String); override;
    Function  GetBody : String; override;

    public
    MajorVersion,
    MinorVersion  : Integer;
    Comment       : String;
  end;

  { TMimeContentTypeField                                                      }
  TMimeContentTypeField = class (TMimeHeaderField)
    protected
    FMajorType,
    FSubType     : String;

    Function  GetAsString : String; override;
    Function  GetName : String; override;
    Procedure SetBody (const FieldBody : String); override;
    Function  GetBody : String; override;
    Function  GetContentType : String;
    Procedure SetContentType (const ContentType : String);
    Function  GetBoundary : String;
    Procedure SetBoundary (const Boundary : String);

    public
    Constructor CreateEx (const MajorType, SubType, Parameters : String); overload;
    Constructor CreateEx (const ContentType : String); overload;
    Property  ContentType : String read GetContentType write SetContentType;
    Property  MajorType : String read FMajorType write FMajorType;
    Property  SubType : String read FSubType write FSubType;
    Property  Boundary : String read GetBoundary write SetBoundary;
    Procedure GenerateBoundary;
  end;

  { TMimeContentTransferEncodingField                                          }
  TMimeTransferEncoding = (te_none, te_7bit, te_8bit, te_binary,
                       te_quoted_printable, te_base64, te_x_uuencode);
  TMimeContentTransferEncodingField = class (THeaderField)
    protected
    FTransferEncoding : TMimeTransferEncoding;

    Function  GetName : String; override;
    Procedure SetBody (const FieldBody : String); override;
    Function  GetBody : String; override;
    Function  GetAsString : String;

    public
    Constructor Create;
    Constructor CreateEx (const TransferEncoding : TMimeTransferEncoding);
    Property TransferEncoding : TMimeTransferEncoding read FTransferEncoding write FTransferEncoding;
  end;



{                                                                              }
{ MIME Message                                                                 }
{                                                                              }
type
  { TMimeMessageHeader                                                         }
  TMimeMessageHeader = class (THeader)
    protected
    Procedure AddField (const FieldName, FieldBody : String); override;
    Function  GetContentTransferEncoding : TMimeTransferEncoding;
    Procedure SetContentTransferEncoding (const ContentTransferEncoding : TMimeTransferEncoding);
    Function  GetContentType : String;
    Procedure SetContentType (const ContentType : String);
    Function  GetContentTypeField : TMimeContentTypeField;

    public
    class Function CreateInstance : AType; override;
    Property  ContentTransferEncoding : TMimeTransferEncoding read GetContentTransferEncoding write SetContentTransferEncoding;
    Property  ContentType : String read GetContentType write SetContentType;
    Property  ContentTypeField : TMimeContentTypeField read GetContentTypeField;
  end;

  { TMimeMessage                                                               }
  TMimeMessage = class (TMessage)
    protected
    Function GetHeader : AHeader; override;
  end;



{                                                                              }
{ MIME E-mail Message                                                          }
{                                                                              }
type
  { TMimeEMailMessageHeader                                                    }
  TMimeEMailMessageHeader = class (TMimeMessageHeader)
    protected
    FIsMimeMessageHeader : Boolean;

    Procedure AddField (const FieldName, FieldBody : String); override;
    Procedure SetAsString (const Header : String); override;

    public
    class Function CreateInstance : AType; override;
    Procedure Prepare; override;
    Property  IsMimeMessageHeader : Boolean read FIsMimeMessageHeader write FIsMimeMessageHeader;
  end;

  { TMimeEMailMessage                                                          }
  TMimeEMailMessage = class (TMimeMessage)
    protected
    FBoundary           : String;
    FIsMimeMessage      : Boolean;
    FPreamble           : String;

    Function  GetAsString : String; override;
    Function  GetHeader : AHeader; override;
    Procedure Init; override;

    public
    Procedure Prepare; override;
    Property  IsMimeMessage : Boolean read FIsMimeMessage write FIsMimeMessage;
    Property  Preamble : String read FPreamble write FPreamble;
  end;

implementation

uses
  // System units
  SysUtils,

  // Delphi Fundamentals
  cUtils,
  cStrings,
  cMaths;



{                                                                              }
{ Encoding / Decoding                                                          }
{                                                                              }
Function DecodeMimeParameterValue (const Parameters, ParameterName : String) : String;
var I, J : Integer;
    L : String;
  Begin
    L := LowerCase (Parameters);
    I := PosNext (LowerCase (ParameterName) + '=', L);
    if I = 0 then
      Result := '' else
      begin
        Inc (I, Length (ParameterName) + 1);
        J := PosNext (';', L, I);
        if J = 0 then
          Result := CopyFrom (Parameters, I) else
          Result := CopyRange (Parameters, I, J - 1);
        Result := RemoveQuotes (Result, '"');
      end;
  End;

Procedure RemoveMimeParameter (var Parameters : String; const ParamName : String);
var S : String;
    I, J : Integer;
  Begin
    S := LowerCase (Parameters);
    I := PosNext (LowerCase (ParamName) + '=', S);
    if I = 0 then
      exit;
    J := PosNext (';', S);
    if J = 0 then
      J := Length (S);
    Delete (Parameters, I, J - I + 1);
  End;

Function EncodeMimeParamValue (const ParamValue : String) : String;
  Begin
    if PosNext ([' ', ';', '='], ParamValue) > 0 then
      Result := '"' + ParamValue + '"' else
      Result := ParamValue;
  End;

Function EncodeMimeParameters (const ParamName, ParamValue : Array of String) : String;
var I, L : Integer;
    S : String;
  Begin
    Assert (High (ParamName) = High (ParamValue), 'EncodeMimeParameters: Must have equal number of entries in parameters.');
    L := High (ParamName);
    Result := '';
    For I := 0 to L do
      if (ParamName [I] <> '') or (ParamValue [I] <> '') then
        begin
          if Result <> '' then
            Result := Result + ';';
          if ParamValue [I] = '' then
            Result := Result + ParamName [I] else
          if ParamName [I] = '' then
            Result := Result + EncodeMimeParamValue (ParamValue [I]) else
            Result := ParamName [I] + '=' + EncodeMimeParamValue (ParamValue [I]);
        end;
  End;

Procedure SetMimeParameter (var Parameters : String; const ParamName, ParamValue : String);
  Begin
    RemoveMimeParameter (Parameters, ParamName);
    Parameters := Parameters + Cond (Parameters <> '', ';', '') +
                  ParamName + '=' + EncodeMimeParamValue (ParamValue);
  End;

Function DecodeMimeFieldBody (const FieldBody : String; var Parameters : String) : String;
var I : Integer;
  Begin
    I := PosNext (';', FieldBody);
    if I = 0 then
      begin
        Result := FieldBody;
        Parameters := '';
      end else
      begin
        Result := CopyLeft (FieldBody, I - 1);
        Parameters := CopyFrom (FieldBody, I + 1);
      end;
  End;

Function EncodeMimeFieldBody (const FieldBody : String; const Parameters : String) : String;
  Begin
    Result := FieldBody + Cond (Parameters <> '', '; ' + Parameters, '');
  End;

Function EncodeMimeFieldBody (const FieldBody : String;
         const ParamName, ParamValue : Array of String) : String;
  Begin
    Result := EncodeMimeFieldBody (FieldBody, EncodeMimeParameters (ParamName, ParamValue));
  End;

Function EncodeMimeHeaderField (const FieldName, FieldBody : String;
         const ParamName, ParamValue : Array of String) : String;
  Begin
    Result := HeaderField (FieldName, EncodeMimeFieldBody (FieldBody, ParamName, ParamValue));
  End;

Function EncodeMimeHeaderField (const FieldName, FieldBody : String; const Parameters : String) : String;
  Begin
    Result := HeaderField (FieldName, FieldBody, Parameters);
  End;

{ Examples from RFC2045:                                                       }
{   "MIME-Version: 1.0                                                         }
{    MIME-Version: 1.0 (produced by MetaSend Vx.x)                             }
{    MIME-Version: (produced by MetaSend Vx.x) 1.0                             }
{    MIME-Version: 1.(produced by MetaSend Vx.x)0                           "  }
Procedure DecodeMimeVersionFieldBody (const FieldBody : String; var MajorVersion, MinorVersion : Integer; var Comment : String);
var B, P, S : String;
  Begin
    B := DecodeMimeFieldBody (FieldBody, P);
    Comment := CopyBetween (B, '(', ')', False);
    S := Remove ('(', ')', B);
    MajorVersion := StrToIntDef (Trim (CopyBefore (B, '.')), 0);
    MinorVersion := StrToIntDef (Trim (CopyAfter (B, '.')), 0);
  End;

Function EncodeMimeVersionFieldBody (const MajorVersion : Integer; const MinorVersion : Integer; const Comment : String) : String;
  Begin
    Result := EncodeMimeFieldBody (IntToStr (MajorVersion) + '.' + IntToStr (MinorVersion) +
              Cond (Comment <> '', ' (' + Comment + ')', ''));
  End;

{ RFC2045:                                                                     }
{   " content := "Content-Type" ":" type "/" subtype                           }
{                *(";" parameter)                                           "  }
Procedure DecodeMimeContentTypeFieldBody (const FieldBody : String; var MajorType, SubType, Parameters : String);
var B : String;
  Begin
    B := DecodeMimeFieldBody (FieldBody, Parameters);
    MajorType := LowerCase (CopyBefore (B, '/', True));
    SubType := LowerCase (CopyAfter (B, '/', False));
  End;

Function EncodeMimeContentTypeFieldBody (const ContentType : String; const Parameters : String) : String;
  Begin
    Result := EncodeMimeFieldBody (ContentType, Parameters);
  End;

Function EncodeMimeContentTypeFieldBody (const MajorType, SubType, Parameters : String) : String;
  Begin
    Result := EncodeMimeContentTypeFieldBody (MajorType + '/' + SubType, Parameters);
  End;

{ Content-Length                                                               }
Function EncodeMimeContentLengthField (const Length : Int64) : String;
  Begin
    Result := HeaderField ('Content-Length', IntToStr (Length));
  End;

{ Mime boundary                                                                }
var
  MIMEBoundaryCounterInit : Boolean = False;
  MIMEBoundaryCounter : Word = 1;

Function GenerateMIMEBoundary : String;
  Begin
    if not MIMEBoundaryCounterInit then
      begin
        MIMEBoundaryCounter := Word (RandomUniform and $FFFF);
        MIMEBoundaryCounterInit := True;
      end else
      Inc (MIMEBoundaryCounter);
    Result := 'df' + LowerCase (IntToHex (MIMEBoundaryCounter) + RandomHex (18));
  End;



{ AMimeHeaderField                                                             }
Function AMimeHeaderField.GetParameters : String;
  Begin
    Result := FParameters;
  End;

Procedure AMimeHeaderField.SetParameters (const Parameters : String);
  Begin
    FParameters := Parameters;
  End;

Function AMimeHeaderField.GetParameter (const ParamName : String) : String;
  Begin
    Result := DecodeMimeParameterValue (GetParameters, ParamName);
  End;

Procedure AMimeHeaderField.SetParameter (const ParamName : String; const ParamValue : String);
var P : String;
  Begin
    P := GetParameters;
    SetMimeParameter (P, ParamName, ParamValue);
    SetParameters (P);
  End;



{ TMimeHeaderField                                                             }
Function TMimeHeaderField.GetParameters : String;
  Begin
    Result := FParameters;
  End;

Procedure TMimeHeaderField.SetParameters (const Parameters : String);
  Begin
    FParameters := Parameters;
  End;

Function TMimeHeaderField.GetParameter (const ParamName : String) : String;
  Begin
    Result := DecodeMimeParameterValue (GetParameters, ParamName);
  End;

Procedure TMimeHeaderField.SetParameter (const ParamName : String; const ParamValue : String);
var P : String;
  Begin
    P := GetParameters;
    SetMimeParameter (P, ParamName, ParamValue);
    SetParameters (P);
  End;



{ TMimeVersionField                                                            }
Procedure TMimeVersionField.SetBody (const FieldBody : String);
  Begin
    DecodeMimeVersionFieldBody (DecodeMimeFieldBody (FieldBody, FParameters), MajorVersion, MinorVersion, Comment);
  End;

Function TMimeVersionField.GetBody : String;
  Begin
    Result := EncodeMimeFieldBody (EncodeMimeVersionFieldBody (MajorVersion, MinorVersion, Comment), FParameters);
  End;

Function TMimeVersionField.GetName : String;
  Begin
    Result := 'MIME-Version';
  End;



{ TMimeContentTypeField                                                        }
Constructor TMimeContentTypeField.CreateEx (const MajorType, SubType, Parameters : String);
  Begin
    inherited Create;
    FMajorType := MajorType;
    FSubType := SubType;
    FParameters := Parameters;
  End;

Constructor TMimeContentTypeField.CreateEx (const ContentType : String);
  Begin
    inherited Create;
    SetContentType (ContentType);
  End;

Procedure TMimeContentTypeField.SetBody (const FieldBody : String);
  Begin
    DecodeMimeContentTypeFieldBody (FieldBody, FMajorType, FSubType, FParameters);
  End;

Function TMimeContentTypeField.GetBody : String;
  Begin
    Result := EncodeMimeContentTypeFieldBody (FMajorType, FSubType, FParameters);
  End;

Function TMimeContentTypeField.GetName : String;
  Begin
    Result := 'Content-Type';
  End;

Function TMimeContentTypeField.GetAsString : String;
  Begin
    if (FMajorType = '') and (FSubType = '') and (FParameters = '') then
      Result := '' else
      Result := inherited GetAsString;
  End;

Function TMimeContentTypeField.GetContentType : String;
  Begin
    if (FMajorType = '') and (FSubType = '') then
      Result := '' else
      Result := FMajorType + '/' + FSubType;
  End;

Procedure TMimeContentTypeField.SetContentType (const ContentType : String);
var P : String;
  Begin
    DecodeMimeContentTypeFieldBody (ContentType, FMajorType, FSubType, P);
  End;

Function TMimeContentTypeField.GetBoundary : String;
  Begin
    Result := DecodeMimeParameterValue (FParameters, 'boundary');
  End;

Procedure TMimeContentTypeField.SetBoundary (const Boundary : String);
  Begin
    SetMimeParameter (FParameters, 'boundary', Boundary);
  End;

Procedure TMimeContentTypeField.GenerateBoundary;
  Begin
    SetMimeParameter (FParameters, 'boundary', GenerateMimeBoundary);
  End;

{ TMimeContentTransferEncodingField                                            }
{ From RFC2045 (Multipurpose Internet Mail Extensions (MIME) Part One: Format  }
{   of Internet Message Bodies)                                                }
{                                                                              }
{         encoding := "Content-Transfer-Encoding" ":" mechanism                }
{         mechanism := "7bit" / "8bit" / "binary" /                            }
{                      "quoted-printable" / "base64" /                         }
{                      ietf-token / x-token                                    }
{         x-token := "x-uuencode"                                              }
Constructor TMimeContentTransferEncodingField.Create;
  Begin
    inherited Create;
    FTransferEncoding := te_None;
  End;

Constructor TMimeContentTransferEncodingField.CreateEx (const TransferEncoding : TMimeTransferEncoding);
  Begin
    inherited Create;
    FTransferEncoding := TransferEncoding;
  End;

Function TMimeContentTransferEncodingField.GetAsString : String;
  Begin
    if FTransferEncoding = te_None then
      Result := '' else
      Result := inherited GetAsString;
  End;

Procedure TMimeContentTransferEncodingField.SetBody (const FieldBody : String);
var B : String;
  Begin
    B := LowerCase (FieldBody);
    if B = '7bit' then
      TransferEncoding := te_7bit else
    if B = '8bit' then
      TransferEncoding := te_8bit else
    if B = 'binary' then
      TransferEncoding := te_binary else
    if B = 'quoted-printable' then
      TransferEncoding := te_quoted_printable else
    if B = 'base64' then
      TransferEncoding := te_base64 else
    if B = 'x-uuencode' then
      TransferEncoding := te_x_uuencode else
      TransferEncoding := te_none;
  End;

Function TMimeContentTransferEncodingField.GetName : String;
  Begin
    Result := 'Content-Transfer-Encoding';
  End;

Function TMimeContentTransferEncodingField.GetBody : String;
  Begin
    Case TransferEncoding of
      te_7bit             : Result := '7bit';
      te_8bit             : Result := '8bit';
      te_binary           : Result := 'binary';
      te_quoted_printable : Result := 'quoted-printable';
      te_base64           : Result := 'base64';
      te_x_uuencode       : Result := 'x-uuencode';
      te_none             : Result := '';
    end;
  End;




{                                                                              }
{ MIME Message                                                                 }
{                                                                              }
class Function TMimeMessageHeader.CreateInstance : AType;
  Begin
    Result := TMimeMessageHeader.Create;
  End;

Procedure TMimeMessageHeader.AddField (const FieldName, FieldBody : String);
var S : String;
    H : CHeaderField;
  Begin
    S := LowerCase (FieldName);
    if S = 'content-type' then
      H := TMimeContentTypeField else
    if S = 'content-transfer-encoding' then
      H := TMimeContentTransferEncodingField else
      H := nil;
    if Assigned (H) then
      AddField (H.CreateEx (FieldBody)) else
      inherited AddField (FieldName, FieldBody);
  End;

Function TMimeMessageHeader.GetContentTransferEncoding : TMimeTransferEncoding;
var F : TMimeContentTransferEncodingField;
  Begin
    F := Field ['content-transfer-encoding'] as TMimeContentTransferEncodingField;
    if not Assigned (F) then
      Result := te_None else
      Result := F.TransferEncoding;
  End;

Procedure TMimeMessageHeader.SetContentTransferEncoding (const ContentTransferEncoding : TMimeTransferEncoding);
  Begin
    Field ['content-transfer-encoding'] := TMimeContentTransferEncodingField.CreateEx (ContentTransferEncoding);
  End;

Function TMimeMessageHeader.GetContentTypeField : TMimeContentTypeField;
  Begin
    Result := Field ['content-type'] as TMimeContentTypeField;
    if not Assigned (Result) then
      begin
        Result := TMimeContentTypeField.CreateEx (ContentType);
        Field ['content-type'] := Result;
      end;
  End;

Function TMimeMessageHeader.GetContentType : String;
var F : TMimeContentTypeField;
  Begin
    F := Field ['content-type'] as TMimeContentTypeField;
    if not Assigned (F) then
      Result := '' else
      Result := F.ContentType;
  End;

Procedure TMimeMessageHeader.SetContentType (const ContentType : String);
  Begin
    GetContentTypeField.ContentType := ContentType;
  End;

Function TMimeMessage.GetHeader : AHeader;
  Begin
    if not Assigned (FHeader) then
      FHeader := TMimeMessageHeader.Create;
    Result := FHeader;
  End;



{                                                                              }
{ MIME E-mail Message                                                          }
{                                                                              }
class Function TMimeEMailMessageHeader.CreateInstance : AType;
  Begin
    Result := TMimeEMailMessageHeader.Create;
  End;

Procedure TMimeEMailMessageHeader.AddField (const FieldName, FieldBody : String);
var S : String;
    H : CHeaderField;
  Begin
    S := LowerCase (FieldName);
    if S = 'mime-version' then
      H := TMimeVersionField else
      H := nil;
    if Assigned (H) then
      AddField (H.CreateEx (FieldBody)) else
      inherited AddField (FieldName, FieldBody);
  End;

Procedure TMimeEMailMessageHeader.Prepare;
  Begin
    if FIsMimeMessageHeader then
      begin
        Field ['MIME-Version'] := TMimeVersionField.CreateEx ('1.0');
      end else
      begin
        Field ['MIME-Version'] := nil;
      end;
    Field ['Date'] := TDateField.CreateNow;
  End;

Procedure TMimeEMailMessageHeader.SetAsString (const Header : String);
var Ver : TMimeVersionField;
    CT  : TMimeContentTypeField;
  Begin
    inherited SetAsString (Header);
    Ver := Field ['MIME-Version'] as TMimeVersionField;
    CT := Field ['Content-Type'] as TMimeContentTypeField;
    FIsMimeMessageHeader := Assigned (Ver) and (Ver.MajorVersion = 1) or
                            Assigned (CT) or
                            (ContentTransferEncoding <> te_none);
  End;

Procedure TMimeEMailMessage.Init;
  Begin
    FIsMimeMessage := True;
  End;

Procedure TMimeEMailMessage.Prepare;
  Begin
    (Header as TMimeEmailMessageHeader).IsMimeMessageHeader := IsMimeMessage;
    inherited Prepare;
  End;

Function TMimeEMailMessage.GetHeader : AHeader;
  Begin
    if not Assigned (FHeader) then
      FHeader := TMimeEmailMessageHeader.Create;
    Result := FHeader;
  End;

Function TMimeEMailMessage.GetAsString : String;
var H : TMimeEMailMessageHeader;
    I : Integer;
  Begin
    H := Header as TMimeEMailMessageHeader;
    Result := Header.AsString +
              Preamble + CRLF;
    For I := 0 to BodyPartCount - 1 do
      Result := Result + '--' + H.ContentTypeField.Boundary + CRLF +
                BodyPart [I].AsString + CRLF;
    Result := Result + '--' + H.ContentTypeField.Boundary + '--' + CRLF;
  End;

end.

