{ Conditional defines:                                                         }
{     ICS - Use Francois Piette's TWSocket instead of Delphi's sockets. (Not   }
{           tested)                                                            }
{$INCLUDE cHeader.inc}
unit cStreams;

{                                                                              }
{                       Streams and Streamers v0.35 (L0)                       }
{                                                                              }
{      This unit is copyright  1999-2001 by David Butler (david@e.co.za)      }
{                                                                              }
{                  This unit is part of Delphi Fundamentals.                   }
{                    Its original file name is cStreams.pas                    }
{                      It was generated 4 Jan 2001 16:42.                      }
{                                                                              }
{                I invite you to use this unit, free of charge.                }
{        I invite you to distibute this unit, but it must be for free.         }
{             I also invite you to contribute to its development,              }
{             but do not distribute a modified copy of this file.              }
{       Send modifications, suggestions and bug reports to david@e.co.za       }
{                                                                              }
{                                                                              }
{ Revision history:                                                            }
{   01/03/99  v0.01   Initial version of TextParser (Skip, Extract and Match   }
{                     methods)                                                 }
{   11/05/99  v0.02   Whitespace functions                                     }
{   20/06/99  v0.03   Added MatchString                                        }
{                     Added Line, Col, LineDelim properties                    }
{                     199 lines interface, 207 lines implementation            }
{   26/06/99  v0.10   Moved AStream from cDataStruct unit.                     }
{                     Added ATBinaryStream. Added ReadPacked functions.        }
{                     Added ATTextStream from TTextParser in unit cTextParser. }
{                     Added Peek, Skip, OctetsRemaing and OctetsAvailable to   }
{                     AStream.                                                 }
{                     Added ExtractPacked functions                            }
{                     Moved from cTextParser and renamed TTextParser to        }
{                     ATTextStream. It now inherits from AStream to allow all  }
{                     derived streams to use it.                               }
{   03/07/99  v0.11   Added MetaCharacter support (machine dependant charsets),}
{                     WhiteSpace functions are now implemented using these.    }
{                     Added MetaString support.                                }
{   17/07/99  v0.12   Added Col property.                                      }
{   03/11/99  v0.13   Revised AStream documentation.                           }
{   04/11/99  v0.14   Overloaded methods in ATBinaryStream. Added Int64.       }
{   05/11/99  v0.15   Removed MetaCharacter support.                           }
{   06/11/99  v0.16   Revisited whole unit.                                    }
{                     Overloaded methods.                                      }
{                     Added general functions to ATStream.                     }
{                     Renamed TStringStream to TMemoryStream.                  }
{                     433 lines interface, 933 lines implementation.           }
{   09/11/99  v0.17   Moved WritePacked's implementations to cStrings unit as  }
{                     functions called Pack.                                   }
{   26/12/99  v0.18   Bug fixes. Added SelfTest procedure.                     }
{   03/01/00  v0.19   Removed FLineDelim. A line delimiter is now defined as   }
{                     EOF or either or both of CR and LF.                      }
{   08/02/00  v0.20   Added ATRecordStream.                                    }
{   27/02/00  v0.21   AStream now derives from Delphi's TStream.               }
{                     Revisited whole unit.                                    }
{                     521 lines interface, 1468 lines implementation.          }
{   05/03/00  v0.22   Renamed TMemoryStream to TStringStream.                  }
{   12/03/00  v0.23   Added TEOLType to ATTextStream.                          }
{   23/04/00  v0.24   Added TSocketStream for use in cHTTPServer.              }
{   08/05/00  v0.25   Cleaned up unit. Merged ATStream, ATBinaryStream,        }
{                     ATTextStream into AStream.                               }
{                     Removed ATRecordStream from TStream and created          }
{                     TFixedLenRecordStreamer.                                 }
{                     405 lines interface, 1307 lines implementation.          }
{   30/05/00  v0.26   Split TStream into TStream (abstract), TExStream         }
{                     (default implementations) and various TStreamer classes, }
{                     including TTextParser (back to v0.01 :-)                 }
{   01/06/00  v0.27   Renamed TFixedLenRecordStreamer to TRecordStreamer.      }
{                     Added TIntegerArrayStreamer from TStreamArray in         }
{                     cDataStructs and added TExtendedArrayStreamer,           }
{                     TShortStringArrayStreamer and TBitArrayStreamer.         }
{   08/06/00  v0.28   Converted TBinaryStreamer methods to plain functions.    }
{                     595 lines interface, 1862 lines implementation.          }
{   14/06/00  v0.29   Converted to template.                                   }
{                     Tmpl: 550 lines interface, 1663 lines implementation.    }
{                     Source: 640 lines interface, 2026 lines implementation.  }
{   11/07/00  v0.30   Moved the bulk of the TTextParser functions out to be    }
{                     stand-alone functions.                                   }
{   21/07/00  v0.31   Added ExtractEmptyLineTerminated.                        }
{   20/08/00  v0.32   Added ExtractDotLineTerminated.                          }
{   21/08/00  v0.33   Bug fixes.                                               }
{   12/09/00  v0.34   Fixed pesky bug in TSocketStream caused by Delphi's      }
{                     TWinSockStream that sometimes "eat" characters.          }
{   01/11/00  v0.35   Some optimizations on TSocketStream.                     }
{                                                                              }
interface

uses
  Windows,              // GetTickCount
  SysUtils,             // Exception, fmxxx constants
  Classes,              // TStream, TFileStream
  WinSock,              // TSocket
  {$IFDEF ICS}
  WSocket,              // Francois Piette's TWSocket
  {$ELSE}
  scktcomp,             // TWinSocketStream
  {$ENDIF}

  // Delphi Fundamentals (L0)
  cUtils;






{                                                                              }
{ TStream                                                                      }
{   Abstract base class for stream implementations.                            }
{                                                                              }
{   EOF is False if more octects may become available.                         }
{                                                                              }
{   Read blocks until EOF or until the data is available. It raises an         }
{   exception if the data can not be read (eg because of a time-out).          }
{                                                                              }
type
  EStream = class (Exception);
  TOctet = Byte;
  TStream = class (Classes.TStream)
    protected
    { Classes.TStream implementations                                          }
    Procedure SetSize (NewSize : Integer); override;
    Function  Seek (Offset : Integer; Origin : Word) : Integer; override;

    { Abstract properties                                                      }
    Function  GetStreamSize : Int64; virtual; abstract;
    Procedure SetStreamSize (const NewSize : Int64); virtual; abstract;
    Function  GetPosition : Int64; virtual; abstract;
    Procedure SetPosition (const NewPosition : Int64); virtual; abstract;

    public
    { Stream attributes                                                        }
    Function  SequentialOnly : Boolean; virtual; abstract;
    Function  SizeUnknown : Boolean; virtual; abstract;

    { Content attributes                                                       }
    Property  Position : Int64 read GetPosition write SetPosition;
    Property  Size : Int64 read GetStreamSize write SetStreamSize;
    Function  Remaining : Int64; virtual; abstract;
    Function  EOF : Boolean; virtual; abstract;

    { Content access                                                           }
    Function  Read (var Buffer; Count : Longint) : Longint; reintroduce; overload; override;
    Function  Write (const Buffer; Count : Longint) : Longint; reintroduce; overload; override;
  end;



{                                                                              }
{ TExStream (Extended Stream)                                                  }
{   Methods have default implementations using above abstract methods.         }
{   Stream implementations should override these with more efficient           }
{   implementations.                                                           }
{                                                                              }
{   Locate returns the position (relative to the current position) of the      }
{   first occurance of Find in the stream. It does not change the current      }
{   position. Returns -1 if it could not be found.                             }
{                                                                              }
type
  TExStream = class;
  TProgressNotifyEvent = Procedure (const Sender : TExStream; const Position, Total : Int64; var Abort : Boolean) of object;
  TExStream = class (TStream)
    protected
    FOnWriteProgress : TProgressNotifyEvent;

    Function  GetAsString : String; virtual;
    Procedure SetAsString (const S : String); virtual;

    public
    Function  Duplicate : TExStream; virtual; abstract;

    { Read/Write                                                               }
    Function  Read : TOctet; overload; virtual;
    Function  Read (const Count : Integer) : String; overload; virtual;
    Procedure Write (const S : TOctet); overload; virtual;
    Procedure Write (const S : String); overload; virtual;
    Procedure Write (const S : Classes.TStream; const BlockSize : Integer); overload; virtual;
    Procedure Write (const S : Classes.TStream; const Count : Int64; const BlockSize : Integer); overload; virtual;
    Property  OnWriteProgress : TProgressNotifyEvent read FOnWriteProgress write FOnWriteProgress;

    Property  AsString : String read GetAsString write SetAsString;
    Procedure Assign (const S : Classes.TStream); virtual;

    Function  Remaining : Int64; override;
    Function  EOF : Boolean; override;

    Procedure Skip; overload; virtual;
    Procedure Skip (const Count : Integer); overload; virtual;

    Function  Peek : TOctet; overload; virtual;
    Function  Peek (const Count : Integer) : String; overload; virtual;

    Procedure Reset; virtual;
    Procedure Truncate; virtual;
    Procedure Append; overload; virtual;

    { Match/Locate                                                             }
    Function  Extract (const S : String) : String; virtual;
    Function  Match (const S : String) : Boolean; overload; virtual;
    Function  Match (const Pos, Length : Int64) : Boolean; overload; virtual;
    Function  Match (const S : Classes.TStream; const Length : Int64) : Boolean; overload; virtual;
    Function  IsEqual (const S : Classes.TStream) : Boolean; overload; virtual;
    Function  IsEqual (const S : String) : Boolean; overload; virtual;
    Function  Locate (const Ch : CharSet) : Int64; overload; virtual;
    Function  Locate (const S : String) : Int64; overload; virtual;

    { Bulk copies                                                              }
    Procedure CopyFrom (const Offset : Int64; const Count : Integer); virtual;
    Procedure Delete (const Count : Integer); virtual;
    Procedure Insert (const Count : Integer); virtual;
  end;



{                                                                              }
{ Binary streamer functions                                                    }
{   Implements functionality to read/write binary data types from.to streams.  }
{                                                                              }
Function  ReadPackedShortInt (const Stream : TExStream) : ShortInt;
Function  ReadPackedWord (const Stream : TExStream) : Word;
Function  ReadPackedSmallInt (const Stream : TExStream) : SmallInt;
Function  ReadPackedLongWord (const Stream : TExStream) : LongWord;
Function  ReadPackedInteger (const Stream : TExStream) : Integer;
Function  ReadPackedInt64 (const Stream : TExStream) : Int64;
Function  ReadPackedBoolean (const Stream : TExStream) : Boolean;
Function  ReadPackedSingle (const Stream : TExStream) : Single;
Function  ReadPackedDouble (const Stream : TExStream) : Double;
Function  ReadPackedExtended (const Stream : TExStream) : Extended;
Function  ReadPackedCurrency (const Stream : TExStream) : Currency;
Function  ReadPackedDateTime (const Stream : TExStream) : TDateTime;
Function  ReadPackedString (const Stream : TExStream) : String;
Function  ReadPackedShortString (const Stream : TExStream) : ShortString;

Procedure WritePackedShortInt (const Stream : TExStream; const D : ShortInt);
Procedure WritePackedWord (const Stream : TExStream; const D : Word);
Procedure WritePackedSmallInt (const Stream : TExStream; const D : SmallInt);
Procedure WritePackedLongWord (const Stream : TExStream; const D : LongWord);
Procedure WritePackedInteger (const Stream : TExStream; const D : Integer);
Procedure WritePackedInt64 (const Stream : TExStream; const D : Int64);
Procedure WritePackedBoolean (const Stream : TExStream; const D : Boolean);
Procedure WritePackedSingle (const Stream : TExStream; const D : Single);
Procedure WritePackedDouble (const Stream : TExStream; const D : Double);
Procedure WritePackedExtended (const Stream : TExStream; const D : Extended);
Procedure WritePackedCurrency (const Stream : TExStream; const D : Currency);
Procedure WritePackedDateTime (const Stream : TExStream; const D : TDateTime);
Procedure WritePackedString (const Stream : TExStream; const D : String);
Procedure WritePackedShortString (const Stream : TExStream; const D : ShortString);



{                                                                              }
{ Text streamer functions                                                      }
{                                                                              }
{   The Skip functions return True if any characters were skipped; and exit    }
{     without raising an exception if EOF is reached.                          }
{   Match functions return True if next characters match, without moving the   }
{     current position. If the match overruns EOF then the function returns    }
{     False.                                                                   }
{   The Extract functions return the extracted text, unchanged. It returns ''  }
{     if a match could not be made.                                            }
{   Function Extract : Char raises an exception if called on EOF.              }
{   ExtractQuotedText extracts text enclosed in a pair Quote.                  }
{     If AllowQuotedQuotes then quotes can be included inside the text by      }
{     using two consequetive quotes, eg for Quote='"', "A""BC" extracts A"BC   }
{     if ExtractLiteral=False and "A""BC" if ExtractLiteral=True.              }
{   MatchKeyword matches from a list of possible keywords, followed by a       }
{     delimiter (if RightDelimiter <> []) or EOF. It returns the index         }
{     of the extracted keyword, or -1 if none matched.                         }
{                                                                              }
Function  Match (const Stream : TExStream; const Ch : Char) : Boolean; overload;
Function  Extract (const Stream : TExStream) : Char; overload;

Function  SkipChar (const Stream : TExStream; const Ch : CharSet) : Boolean;
Function  MatchChar (const Stream : TExStream; const Ch : CharSet) : Boolean;
Function  ExtractChar (const Stream : TExStream; const Ch : CharSet) : String;

Function  SkipChars (const Stream : TExStream; const Ch : CharSet) : Boolean;
Function  ExtractChars (const Stream : TExStream; const Ch : CharSet) : String;

Function  SkipWhiteSpace (const Stream : TExStream) : Boolean;
Function  MatchWhiteSpace (const Stream : TExStream) : Boolean;
Function  ExtractWhiteSpace (const Stream : TExStream) : String;

Function  SkipTo (const Stream : TExStream; const Delimiter : String) : Boolean;
Function  ExtractTo (const Stream : TExStream; const Delimiter : String; const SkipDelimiter : Boolean = False) : String; overload;
Function  ExtractTo (const Stream : TExStream; const Delimiter : CharSet; const SkipDelimiter : Boolean = False) : String; overload;
Function  ExtractToEOF (const Stream : TExStream) : String;

Function  ExtractQuoted (const Stream : TExStream; const Quote : Char; const AllowQuotedQuote : Boolean) : String;

type
  TEOL = (eol_CR, eol_LF, eol_CRLF, eol_LFCR, eol_EOF);
  TEOLType = Set of TEOL;

Function  MatchEOL (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : Boolean;
Function  SkipEOL (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : Boolean;
Function  ExtractEOL (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : String;
Function  LocateEOL (const Stream : TExStream; const EOLType : TEOLType; var EOLEncountered : TEOL) : Int64;

Function  SkipLine (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : Boolean;
Function  MatchLine (const Stream : TExStream; const S : String; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : Boolean;
Function  ExtractLine (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]; const MaxLineLength : Integer = -1) : String;

Function  MatchKeyword (const Stream : TExStream; const Keywords : StringArray; const RightDelimiter : CharSet) : Integer;
Function  ExtractKeyword (const Stream : TExStream; const Keywords : StringArray; const RightDelimiter : CharSet) : String;

{ Multi-line functions                                                         }
type
  TStreamProgressCallback = Procedure (const Complete : Int64; var Abort : Boolean) of object;

Procedure StreamEmptyLineTerminated (const Source, Dest : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]); overload;
Procedure StreamEmptyLineTerminated (const Source : String; const Dest : TExStream); overload;
Function  EncodeEmptyLineTerminated (const S : String) : String;
Function  ExtractEmptyLineTerminated (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : String;

Procedure StreamDotLineTerminated (const Source, Dest : TExStream;
          const EOLType : TEOLType = [eol_EOF, eol_CRLF];
          const ProgressCallback : TStreamProgressCallback = nil); overload;
Procedure StreamDotLineTerminated (const Source : String; const Dest : TExStream;
          const ProgressCallback : TStreamProgressCallback = nil); overload;
Function  EncodeDotLineTerminated (const S : String) : String;
Function  ExtractDotLineTerminated (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : String;
Function  ExtractMultiLine (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : StringArray;


{                                                                              }
{ TFileStream                                                                  }
{                                                                              }
type
  TFileStreamMode = Set of (fsmCreate,
                            fsmCreateIfNotExist, 
                            fsmDeleteOnClose,
                            fsmReadOnlyAccess,
                            fsmShareAllowWrite,
                            fsmShareDenyRead);
type
  TFileStream = class (TExStream)
    protected
    FFileStream    : Classes.TFileStream;
    FFileCreated   : Boolean;
    FDeleteOnClose : Boolean;
    FFilePath      : String;
    FFileMode      : TFileStreamMode;

    Function  GetStreamSize : Int64; override;
    Procedure SetStreamSize (const NewSize : Int64); override;
    Function  GetFileDate : TDateTime;
    Procedure SetFileDate (const D : TDateTime);

    public
    Constructor Create (const FilePath : String; const Mode : TFileStreamMode);
    Destructor Destroy; override;

    Function  Duplicate : TExStream; override;
    Function  SequentialOnly : Boolean; override;
    Function  SizeUnknown : Boolean; override;
    Function  Read : TOctet; overload; override;
    Function  Read (const Count : Integer) : String; overload; override;
    Function  Read (var Buf; Count : Integer) : Integer; overload; override;
    Function  Write (const Buf; Count : Integer) : LongInt; overload; override;
    Procedure Write (const S : TOctet); overload; override;
    Procedure Write (const S : String); overload; override;
    Procedure Skip (const Count : Integer); overload; override;
    Procedure Skip; overload; override;
    Function  Peek : TOctet; overload; override;
    Function  Peek (const Count : Integer) : String; overload; override;
    Function  GetPosition : Int64; override;
    Procedure SetPosition (const NewPosition : Int64); override;
    Procedure Append; override;
    Function  Remaining : Int64; override;
    Function  EOF : Boolean; override;
    Procedure Truncate; override;
    Function  Locate (const Ch : CharSet) : Int64; overload; override;
    Function  Locate (const S : String) : Int64; overload; override;

    Property  DeleteOnClose : Boolean read FDeleteOnClose write FDeleteOnClose;
    Property  FileDate : TDateTime read GetFileDate write SetFileDate;
  end;



{                                                                              }
{ TStringStream                                                                }
{   A implementation of a stream that stores/retrieves data using a Delphi     }
{   String.                                                                    }
{                                                                              }
type
  TStringStream = class (TExStream)
    protected
    FData : String;
    FPos  : Integer;

    Function GetStreamSize : Int64; override;
    Procedure SetStreamSize (const NewSize : Int64); override;

    public
    Constructor CreateEx (const S : String);

    Function  Duplicate : TExStream; override;
    Procedure Skip (const Count : Integer); overload; override;
    Procedure Skip; overload; override;
    Function  Read : TOctet; overload; override;
    Function  Read (const Count : Integer) : String; overload; override;
    Function  Read (var Buf; Count : Integer) : Integer; override;
    Function  Peek : TOctet; overload; override;
    Function  Peek (const Count : Integer) : String; overload; override;
    Function  GetPosition : Int64; override;
    Procedure SetPosition (const NewPosition : Int64); override;
    Procedure Append; override;
    Function  Remaining : Int64; override;
    Function  EOF : Boolean; override;
    Procedure Truncate; override;
    Function  Write (const Buf; Count : Integer) : LongInt; overload; override;
    Procedure Write (const Data : TOctet); overload; override;
    Procedure Write (const Data : String); overload; override;
    Function  Locate (const S : String) : Int64; overload; override;
    Function  Locate (const Ch : CharSet) : Int64; overload; override;
    Function  SequentialOnly : Boolean; override;
    Function  SizeUnknown : Boolean; override;
    Function  Match (const S : String) : Boolean; override;
  end;



{                                                                              }
{ TSocketStream                                                                }
{                                                                              }
type
  ESocketStream = class (Exception);
  TSocketStream = class (TExStream)
    protected
    FSocketHandle : TSocket;
    {$IFDEF ICS}
    FSocket       : TWSocket;
    {$ELSE}
    FClient       : TClientSocket;
    FSocket       : TCustomWinSocket;
    FSockStream   : TWinSocketStream;
    {$ENDIF}
    FBytesRead    : Integer;
    FPeekBuf      : String;
    FPeekBufPos   : Integer;
    FTimeOut      : Integer;
    FCloseSocket  : Boolean;

    Function  PeekBufLeft : Integer;
    Function  FillPeekBuf (const Len : Integer) : Boolean;
    Function  GetLocalHost : String;

    public
    Constructor Create (const Host : String; const Port : Integer;
                const TimeOut : Integer; const CloseSocketOnDestroy : Boolean = True);
    Constructor CreateEx (const Socket : {$IFDEF ICS}TSocket{$ELSE}TCustomWinSocket{$ENDIF};
                const TimeOut : Integer;
                const CloseSocketOnDestroy : Boolean = True);
    Destructor Destroy; override;

    Function  SizeUnknown : Boolean; override;
    Procedure Skip (const Count : Integer); overload; override;
    Procedure Skip; overload; override;
    Function  Read (var Buf; Count : Integer) : Integer; override;
    Function  Read : TOctet; overload; override;
    Function  Read (const Count : Integer) : String; overload; override;
    Function  ReadAvailable : String; overload;
    Function  Peek : TOctet; overload; override;
    Function  Peek (const Count : Integer) : String; overload; override;
    Function  Match (const S : String) : Boolean; override;
    Function  Locate (const S : String) : Int64; override;
    Function  Locate (const Ch : CharSet) : Int64; override;
    Function  GetPosition : Int64; override;
    Procedure SetPosition (const NewPosition : Int64); override;
    Procedure Append; override;
    Function  EOF : Boolean; override;
    Procedure Write (const Data : TOctet); overload; override;
    Procedure Write (const Data : String); overload; override;
    Function  Write (const Buf; Count : Integer) : LongInt; overload; override;

    Procedure CloseSocket;
    Property  CloseSocketOnDestroy : Boolean read FCloseSocket write FCloseSocket;
    Property  Socket : TSocket read FSocketHandle;
    Property  LocalHost : String read GetLocalHost;
  end;




{                                                                              }
{ TStreamer                                                                    }
{                                                                              }
type
  TStreamer = class
    protected
    FStream : TExStream;

    Procedure Init (const Stream : TExStream); virtual;

    public
    Constructor Create (const Stream : TExStream);
    Destructor Destroy; override;
    Procedure ReleaseStream; virtual;
    Procedure Reset; virtual;
    Property  Stream : TExStream read FStream;
  end;



{                                                                              }
{ Text parser                                                                  }
{                                                                              }
type
  TTextParser = class (TStreamer)
    protected
    FLineNr          : Integer;
    FLastNewLine     : Integer;
    FEOLType         : TEOLType;
    FWhiteSpace      : CharSet;
    FCaseSensitive   : Boolean;
    FLocaleSensitive : Boolean;
    FMaxLineLength   : Integer;

    Function  GetCol : Integer;

    public
    Constructor Create (const Stream : TExStream);

    Property  CaseSensitive : Boolean read FCaseSensitive write FCaseSensitive;
    Property  LocaleSensitive : Boolean read FLocaleSensitive write FLocaleSensitive;
    Property  WhiteSpace : CharSet read FWhiteSpace write FWhiteSpace;
    Property  EOLType : TEOLType read FEOLType write FEOLType;
    Property  Line : Integer read FLineNr;
    Property  Col : Integer read GetCol;
    Property  MaxLineLength : Integer read FMaxLineLength write FMaxLineLength;

    Function  GetBookmark : Int64;
    Procedure GotoBookmark (const B : Int64);
    Procedure Writeln (const S : String);
    Function  Readln : String;
    Function  EOF : Boolean;

    Function  Match (const Txt : String) : Boolean;
    Function  Extract (const Txt : String) : String;
    Function  MatchLine (const Txt : String) : Boolean;
  end;



{                                                                              }
{ TRecordStreamer                                                              }
{   Implements functionality to access fixed-length records in streams.        }
{                                                                              }
type
  TRecordStreamer = class (TStreamer)
    protected
    FRecordSize   : Integer;
    FRecordOffset : Int64;

    Function  GetRecordCount : Int64;
    Procedure SetRecordCount (const NewRecordCount : Int64);
    Procedure SetRecordSize (const RecordSize : Integer);
    Procedure Seek (const Idx : Int64);

    public
    Constructor Create (const Stream : TExStream; const RecordSize : Integer = 1;
                const FirstRecordOffset : Int64 = 0);

    Procedure Read (const Idx : Int64; const Count : Integer; var Buf);
    Procedure Write (const Idx : Int64; const Count : Integer; const Buf);
    Procedure Insert (const Idx : Int64; const Count : Integer);
    Procedure Delete (const Idx : Int64; const Count : Integer);
    Procedure Clear (const Idx : Int64; const Count : Integer);

    Property RecordSize : Integer read FRecordSize write FRecordSize;
    Property RecordOffset : Int64 read FRecordOffset write FRecordOffset;
    Property RecordCount : Int64 read GetRecordCount write SetRecordCount;
  end;

  TIntegerArrayStreamer = class (TRecordStreamer)
    protected
    Procedure SetItem (const Idx : Integer; const Value : Integer);
    Function  GetItem (const Idx : Integer) : Integer;
    Function  GetAsIntegerArray : IntegerArray;
    Procedure SetAsIntegerArray (const V : IntegerArray);

    public
    Constructor Create (const Stream : TExStream);

    Property  Item [const Idx : Integer] : Integer read GetItem write SetItem; default;
    Property  Count : Int64 read GetRecordCount write SetRecordCount;
    Procedure Fill (const Value : Integer);
    Procedure FillInc (const StartValue : Integer = 0; const Increment : Integer = 1);
    Function  PosNext (const Find : Integer; const PrevPos : Integer = -1;
              const IsSortedAscending : Boolean = False) : Integer;
    Property  AsIntegerArray : IntegerArray read GetAsIntegerArray write SetAsIntegerArray;
  end;


  TInt64ArrayStreamer = class (TRecordStreamer)
    protected
    Procedure SetItem (const Idx : Integer; const Value : Int64);
    Function  GetItem (const Idx : Integer) : Int64;
    Function  GetAsInt64Array : Int64Array;
    Procedure SetAsInt64Array (const V : Int64Array);

    public
    Constructor Create (const Stream : TExStream);

    Property  Item [const Idx : Integer] : Int64 read GetItem write SetItem; default;
    Property  Count : Int64 read GetRecordCount write SetRecordCount;
    Procedure Fill (const Value : Int64);
    Procedure FillInc (const StartValue : Int64 = 0; const Increment : Int64 = 1);
    Function  PosNext (const Find : Int64; const PrevPos : Integer = -1;
              const IsSortedAscending : Boolean = False) : Integer;
    Property  AsInt64Array : Int64Array read GetAsInt64Array write SetAsInt64Array;
  end;


  TSingleArrayStreamer = class (TRecordStreamer)
    protected
    Procedure SetItem (const Idx : Integer; const Value : Single);
    Function  GetItem (const Idx : Integer) : Single;
    Function  GetAsSingleArray : SingleArray;
    Procedure SetAsSingleArray (const V : SingleArray);

    public
    Constructor Create (const Stream : TExStream);

    Property  Item [const Idx : Integer] : Single read GetItem write SetItem; default;
    Property  Count : Int64 read GetRecordCount write SetRecordCount;
    Procedure Fill (const Value : Single);
    Procedure FillInc (const StartValue : Single = 0; const Increment : Single = 1);
    Function  PosNext (const Find : Single; const PrevPos : Integer = -1;
              const IsSortedAscending : Boolean = False) : Integer;
    Property  AsSingleArray : SingleArray read GetAsSingleArray write SetAsSingleArray;
  end;


  TDoubleArrayStreamer = class (TRecordStreamer)
    protected
    Procedure SetItem (const Idx : Integer; const Value : Double);
    Function  GetItem (const Idx : Integer) : Double;
    Function  GetAsDoubleArray : DoubleArray;
    Procedure SetAsDoubleArray (const V : DoubleArray);

    public
    Constructor Create (const Stream : TExStream);

    Property  Item [const Idx : Integer] : Double read GetItem write SetItem; default;
    Property  Count : Int64 read GetRecordCount write SetRecordCount;
    Procedure Fill (const Value : Double);
    Procedure FillInc (const StartValue : Double = 0; const Increment : Double = 1);
    Function  PosNext (const Find : Double; const PrevPos : Integer = -1;
              const IsSortedAscending : Boolean = False) : Integer;
    Property  AsDoubleArray : DoubleArray read GetAsDoubleArray write SetAsDoubleArray;
  end;


  TExtendedArrayStreamer = class (TRecordStreamer)
    protected
    Procedure SetItem (const Idx : Integer; const Value : Extended);
    Function  GetItem (const Idx : Integer) : Extended;
    Function  GetAsExtendedArray : ExtendedArray;
    Procedure SetAsExtendedArray (const V : ExtendedArray);

    public
    Constructor Create (const Stream : TExStream);

    Property  Item [const Idx : Integer] : Extended read GetItem write SetItem; default;
    Property  Count : Int64 read GetRecordCount write SetRecordCount;
    Procedure Fill (const Value : Extended);
    Procedure FillInc (const StartValue : Extended = 0; const Increment : Extended = 1);
    Function  PosNext (const Find : Extended; const PrevPos : Integer = -1;
              const IsSortedAscending : Boolean = False) : Integer;
    Property  AsExtendedArray : ExtendedArray read GetAsExtendedArray write SetAsExtendedArray;
  end;


  TShortStringArrayStreamer = class (TRecordStreamer)
    protected
    FStringLen : Byte;

    Procedure SetItem (const Idx : Integer; const Value : ShortString);
    Function  GetItem (const Idx : Integer) : ShortString;

    public
    Constructor Create (const Stream : TExStream; const StringLen : Byte = 255);

    Property  StringLen : Byte read FStringLen;
    Property  Item [const Idx : Integer] : ShortString read GetItem write SetItem; default;
    Property  Count : Int64 read GetRecordCount write SetRecordCount;
    Procedure Fill (const Value : ShortString = '');
    Function  PosNext (const Find : ShortString; const PrevPos : Integer = -1;
              const IsSortedAscending : Boolean = False) : Integer;
  end;

  TBitArrayStreamer = class (TRecordStreamer)
    protected
    Procedure SetItem (const Idx : Int64; const Value : Boolean);
    Function  GetItem (const Idx : Int64) : Boolean;
    Procedure SetCount (const NewCount : Int64);
    Function  GetCount : Int64;

    public
    Constructor Create (const Stream : TExStream);

    Property  Item [const Idx : Int64] : Boolean read GetItem write SetItem; default;
    Property  Count : Int64 read GetCount write SetCount;
    Procedure Fill (const Value : Boolean = False);
    Procedure FillRange (const LoIdx, HiIdx : Int64; const Value : Boolean);
    Function  CompareRange (const LoIdx, HiIdx : Int64; const Value : Boolean) : Boolean;
    Procedure Invert;
    Function  Find (const Value : Boolean; const Start : Integer = 0;
              const FindForward : Boolean = True) : Integer;
    Function  FindRange (const Value : Boolean; const Start : Integer = 0;
              const Count : Integer = 1; const FindForward : Boolean = True) : Integer;
  end;



implementation



uses
  // System units
  Math,                 // Max

  // Delphi Fundamentals (L0)
  cDateTime,            // HP Timer
  cStrings;             // Pack



{                                                                              }
{ TStream                                                                      }
{                                                                              }
Function TStream.Read (var Buffer; Count : Longint) : Longint;
  Begin
    raise EAbstractError.Create ('Method TStream.Read not implemented.');
  End;

Function TStream.Write (const Buffer; Count : Longint) : Longint;
  Begin
    raise EAbstractError.Create ('Method TStream.Write not implemented.');
  End;

Function TStream.Seek (Offset : Integer; Origin : Word) : Integer;
  Begin
    Case Origin of
      soFromBeginning : if Offset >= 0 then
                          SetPosition (Offset) else
                          raise EStream.Create ('Seek before bof');
      soFromCurrent   : SetPosition (GetPosition + Offset);
      soFromEnd       : SetPosition (GetStreamSize + Offset);
    end;
    Result := GetPosition;
  End;

Procedure TStream.SetSize (NewSize : Integer);
  Begin
    SetStreamSize (NewSize);
  End;



{                                                                              }
{ TExStream                                                                    }
{                                                                              }
Function TExStream.Read : TOctet;
  Begin
    if Read (Result, Sizeof (Result)) <> Sizeof (Result) then
      raise EReadError.Create ('Stream read error.');
  End;

Function TExStream.Read (const Count : Integer) : String;
  Begin
    SetLength (Result, Count);
    if Count > 0 then
      SetLength (Result, Read (Result [1], Count));
  End;

Procedure TExStream.Write (const S : TOctet);
  Begin
    if Write (S, Sizeof (S)) <> Sizeof (S) then
      raise EWriteError.Create ('Stream write error.');
  End;

Procedure TExStream.Write (const S : String);
var L : Integer;
  Begin
    L := Length (S);
    if L > 0 then
      if Write (S [1], L) <> L then
        raise EWriteError.Create ('Stream write error.');
  End;

Procedure TExStream.Write (const S : Classes.TStream; const Count : Int64; const BlockSize : Integer);
var Buffer  : String;
    F, P, L : Int64;
    Aborted : Boolean;
  Begin
    Assert (BlockSize > 0, 'TExStream.Write: Invalid BlockSize');
    Aborted := False;
    P := 0;
    L := Min (Count, S.Size - S.Position);
    Repeat
      F := Min (BlockSize, L - P);
      if F > 0 then
        begin
          SetLength (Buffer, F);
          S.ReadBuffer (Buffer [1], F);
          Write (Buffer);
          Inc (P, F);
          if Assigned (FOnWriteProgress) then
            FOnWriteProgress (self, P, L, Aborted);
        end;
    Until (F = 0) or Aborted;
    if Aborted then
      raise EWriteError.Create ('Write operation aborted.');
  End;

Procedure TExStream.Write (const S : Classes.TStream; const BlockSize : Integer);
  Begin
    S.Seek (0, soFromBeginning);
    Write (S, S.Size, BlockSize);
  End;

Procedure TExStream.Assign (const S : Classes.TStream);
  Begin
    Position := 0;
    Write (S, 1024);
    Truncate;
  End;

Function TExStream.Remaining : Int64;
  Begin
    Result := Size - Position;
  End;

Function TExStream.EOF : Boolean;
  Begin
    Result := Size = Position;
  End;

Procedure TExStream.Truncate;
  Begin
    Size := Position;
  End;

Procedure TExStream.Append;
  Begin
    Position := Size;
  End;

Procedure TExStream.Skip;
  Begin
    if SequentialOnly then
      Read else
      Position := Position + 1;
  End;

Procedure TExStream.Skip (const Count : Integer);
  Begin
    if SequentialOnly then
      Read (Count) else
      Position := Min (Size, Position + Count);
  End;

Function TExStream.GetAsString : String;
  Begin
    Position := 0;
    Result := Read (Size);
  End;

Procedure TExStream.SetAsString (const S : String);
  Begin
    Position := 0;
    Write (S);
    Truncate;
  End;

Procedure TExStream.Reset;
  Begin
    Position := 0;
  End;

Function TExStream.Peek : TOctet;
var P : Int64;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Peek not supported.');

    P := Position;
    Result := Read;
    Position := P;
  End;

Function TExStream.Peek (const Count : Integer) : String;
var P : Int64;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Peek not supported.');

    P := Position;
    Result := Read (Count);
    Position := P;
  End;

Function TExStream.Extract (const S : String) : String;
const CompareBufferSize = 1024;
var I, R, L : Integer;
    P       : Int64;
    Buffer  : String;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Match not supported.');

    Result := '';
    if Remaining < Length (S) then
      exit;

    P := Position;
    L := Length (S);
    R := L;
    While R > 0 do
      begin
        I := Min (R, CompareBufferSize);
        Buffer := Read (I);
        if not cStrings.Match (Buffer, S, L - R + 1) then
          begin
            Position := P;
            exit;
          end;
        Dec (R, I);
      end;

    Result := S;
  End;

Function TExStream.Match (const S : String) : Boolean;
var P : Int64;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Match not supported.');

    P := Position;
    Result := Extract (S) <> '';
    if Result then
      Position := P;
  End;

Function TExStream.Match (const Pos, Length : Int64) : Boolean;
const CompareBufferSize = 1024;
var I, R, P : Int64;
    Buffer  : String;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Match not supported.');

    Result := False;
    if Remaining < Length then
      exit;

    R := Length;
    P := Position;
    While R > 0 do
      begin
        I := Min (R, CompareBufferSize);
        Buffer := Read (I);
        Position := Pos + Length - R;
        if Buffer <> Read (I) then
          begin
            Position := P;
            exit;
          end;
        Dec (R, I);
        Position := P + Length - R;
      end;

    Position := P;
    Result := True;
  End;

Function TExStream.Match (const S : Classes.TStream; const Length : Int64) : Boolean;
const CompareBufferSize = 1024;
var P, I, R : Int64;
    Buffer  : String;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Match not supported.');

    Result := False;
    if (Remaining < Length) or (S.Size - S.Position < Length) then
      exit;

    P := Position;
    R := Length;
    While R > 0 do
      begin
        I := Min (R, CompareBufferSize);
        SetLength (Buffer, I);
        S.ReadBuffer (Buffer [1], I);
        if Read (I) <> Buffer then
          begin
            Position := P;
            exit;
          end;
        Dec (R, I);
      end;

    Position := P;
    Result := True;
  End;

Function TExStream.IsEqual (const S : Classes.TStream) : Boolean;
var L1, L2 : Int64;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Match not supported.');

    L1 := Size;
    L2 := S.Size;
    if (L1 >= 0) and (L2 >= 0) and (L1 <> L2) then
      Result := False else
      begin
        Position := 0;
        S.Seek (0, soFromBeginning);
        Result := Match (S, L1);
      end;
  End;

Function TExStream.IsEqual (const S : String) : Boolean;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Match not supported.');

    if Length (S) <> Remaining then
      Result := False else
      begin
        Position := 0;
        Result := Match (S);
      end;
  End;

Function TExStream.Locate (const Ch : CharSet) : Int64;
var P : Int64;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Locate not supported.');

    P := Position;
    While not EOF do
      if Char (Read) in Ch then
        begin
          Result := Position - P;
          Position := P;
          exit;
        end;
    Position := P;
    Result := -1;
  End;

Function TExStream.Locate (const S : String) : Int64;
var P, I : Int64;
    L    : Integer;
  Begin
    if SequentialOnly then
      raise EStream.Create ('Locate not supported.');

    L := Length (S);
    P := Position;
    I := 0;
    While not EOF do
      begin
        if Peek (L) = S then
          begin
            Position := P;
            Result := I;
            exit;
          end;
        Inc (I);
        Position := P + I;
      end;
    Position := P;
    Result := -1;
  End;

Procedure TExStream.CopyFrom (const Offset : Int64; const Count : Integer);
const BufSize = 16384;
var S, D   : Int64;
    L, C   : Integer;
    Buffer : String;
  Begin
    if SequentialOnly then
      raise EStream.Create ('CopyFrom not supported.');

    S := Offset;
    D := Position;
    L := Min (Count, Size - S);
    While L > 0 do
      begin
        C := Min (BufSize, L);
        SetLength (Buffer, C);
        Position := S;
        Read (Buffer [1], C);
        Position := D;
        Write (Buffer [1], C);
        Dec (L, C);
      end;
  End;

Procedure TExStream.Delete (const Count : Integer);
var D, S : Int64;
  Begin
    D := Position + Count;
    S := Size;
    CopyFrom (D, S - D);
    Position := S - Count;
    Truncate;
  End;

Procedure TExStream.Insert (const Count : Integer);
var P : Int64;
  Begin
    P := Position;
    Position := P + Count;
    CopyFrom (P, Size - P);
  End;



{                                                                              }
{ TFileStream                                                                  }
{                                                                              }
Constructor TFileStream.Create (const FilePath : String; const Mode : TFileStreamMode);
var M : Word;
    CreateFile : Boolean;
  Begin
    inherited Create;

    FFilePath := FilePath;
    FFileMode := Mode;

    FDeleteOnClose := fsmDeleteOnClose in Mode;

    CreateFile := ([fsmCreate, fsmCreateIfNotExist] * Mode <> []) and not FileExists (FilePath);

    if CreateFile then
      M := fmCreate else
      if fsmReadOnlyAccess in Mode then
        M := fmOpenRead else
        M := fmOpenReadWrite;
    M := M or fmShareCompat;
    if not (fsmShareAllowWrite in Mode) then
      M := M or fmShareDenyWrite;
    if fsmShareDenyRead in Mode then
      M := M or fmShareDenyRead;

    FFileStream := Classes.TFileStream.Create (FilePath, M);
    if not CreateFile and (fsmCreate in Mode) then
      FFileStream.Size := 0;
    FFileCreated := CreateFile;
  End;

Destructor TFileStream.Destroy;
  Begin
    FreeAndNil (FFileStream);
    if FDeleteOnClose then
      DeleteFile (FFilePath);
    inherited Destroy;
  End;

Function TFileStream.Duplicate : TExStream;
  Begin
    Result := TFileStream.Create (FFilePath, FFileMode);
  End;

Function TFileStream.SequentialOnly : Boolean;
  Begin
    Result := False;
  End;

Function TFileStream.SizeUnknown : Boolean;
  Begin
    Result := False;
  End;

Function TFileStream.Read : TOctet;
  Begin
    FFileStream.ReadBuffer (Result, Sizeof (Result));
  End;

Function TFileStream.Read (var Buf; Count : Integer) : Integer;
  Begin
    if Count <= 0 then
      Result := 0 else
      Result := FFileStream.Read (Buf, Min (Count, Remaining));
  End;

Function TFileStream.Read (const Count : Integer) : String;
var I : Integer;
  Begin
    if Count <= 0 then
      Result := '' else
      begin
        I := Min (Count, Remaining);
        SetLength (Result, I);
        if I > 0 then
          SetLength (Result, FFileStream.Read (Result [1], I));
      end;
  End;

Procedure TFileStream.Write (const S : TOctet);
  Begin
    FFileStream.WriteBuffer (S, Sizeof (S));
  End;

Function TFileStream.Write (const Buf; Count : Integer) : LongInt;
  Begin
    FFileStream.WriteBuffer (Buf, Count);
    Result := Count;
  End;

Procedure TFileStream.Write (const S : String);
var L : Integer;
  Begin
    L := Length (S);
    if L > 0 then
      FFileStream.WriteBuffer (S [1], L);
  End;

Function TFileStream.Remaining : Int64;
  Begin
    Result := FFileStream.Size - FFileStream.Position;
  End;

Function TFileStream.EOF : Boolean;
  Begin
    Result := not (FFileStream.Position < FFileStream.Size);
  End;

Procedure TFileStream.Skip;
  Begin
    if EOF then
      raise EStream.Create ('Skip past EOF');
    FFileStream.Seek (FFileStream.Position + 1, soFromBeginning);
  End;

Procedure TFileStream.Skip (const Count : Integer);
  Begin
    FFileStream.Seek (Min (FFileStream.Position + Count, FFileStream.Size), soFromBeginning);
  End;

Function TFileStream.GetPosition : Int64;
  Begin
    Result := FFileStream.Position;
  End;

Function TFileStream.GetStreamSize : Int64;
  Begin
    Result := FFileStream.Size;
  End;

Procedure TFileStream.SetStreamSize (const NewSize : Int64);
  Begin
    FFileStream.Size := NewSize;
  End;

Procedure TFileStream.SetPosition (const NewPosition : Int64);
  Begin
    FFileStream.Position := NewPosition;
  End;

Function TFileStream.Peek : TOctet;
var P : Integer;
  Begin
    P := FFileStream.Position;
    FFileStream.Read (Result, Sizeof (Result));
    FFileStream.Seek (P, soFromBeginning);
  End;

Function TFileStream.Peek (const Count : Integer) : String;
var P : Integer;
  Begin
    P := FFileStream.Position;
    Result := Read (Count);
    FFileStream.Seek (P, soFromBeginning);
  End;

Procedure TFileStream.Truncate;
  Begin
    FFileStream.Size := FFileStream.Position;
  End;

Procedure TFileStream.Append;
  Begin
    FFileStream.Seek (0, soFromEnd);
  End;

Function TFileStream.Locate (const S : String) : Int64;

  Function ReadBuf (const Len : Integer) : String;
    Begin
      SetLength (Result, Len);
      SetLength (Result, FFileStream.Read (Result [1], Len));
    End;

const BufSize = 1024;

var Buf, B : String;
    P, I   : Integer;

  Begin
    if S = '' then
      exit;
    P := FFileStream.Position;
    try
      Buf := ReadBuf (Length (S));
      Result := PosNext (S, Buf);
      if Result > 0 then
        exit;
      I := 0;
      While FFileStream.Position < FFileStream.Size do
        begin
          B := ReadBuf (BufSize);
          Buf := Buf + B;
          Result := I + PosNext (S, Buf);
          if Result > I then
            exit;
          System.Delete (Buf, 1, Length (B));
          Inc (I, Length (B));
        end;

      Result := -1;
    finally
      FFileStream.Seek (P, soFromBeginning);
    end;
  End;

Function TFileStream.Locate (const Ch : CharSet) : Int64;
const BufSize = 1024;
var Buf  : String;
    P, I : Integer;
  Begin
    P := FFileStream.Position;
    try
      While not EOF do
        begin
          I := Min (Remaining, BufSize);
          SetLength (Buf, I);
          SetLength (Buf, FFileStream.Read (Buf [1], I));

          I := PosNext (Ch, Buf);
          if I > 0 then
            begin
              Result := FFileStream.Position - Length (Buf) + I - 1;
              exit;
            end;
        end;
      Result := -1;
    finally
      FFileStream.Seek (P, soFromBeginning);
    end;
  End;

Function TFileStream.GetFileDate : TDateTime;
  Begin
    Result := FileDateToDateTime (FileGetDate (FFileStream.Handle));
  End;

Procedure TFileStream.SetFileDate (const D : TDateTime);
  Begin
    FileSetDate (FFileStream.Handle, DateTimeToFileDate (D));
  End;



{                                                                              }
{ TStringStream                                                                }
{                                                                              }
Constructor TStringStream.CreateEx (const S : String);
  Begin
    inherited Create;
    FData := S;
  End;

Function TStringStream.SequentialOnly : Boolean;
  Begin
    Result := False;
  End;

Function TStringStream.SizeUnknown : Boolean;
  Begin
    Result := False;
  End;

Function TStringStream.Duplicate : TExStream;
  Begin
    Result := TStringStream.CreateEx (FData);
  End;

Function TStringStream.Read : TOctet;
  Begin
    if FPos = Length (FData) then
      raise EStream.Create ('Read past EOF');
    Inc (FPos);
    Result := TOctet (FData [FPos]);
  End;

Function TStringStream.Read (var Buf; Count : Integer) : Integer;
var L : Integer;
  Begin
    L := Length (FData);
    if FPos + Count > L then
      Result := L - FPos else
      Result := Count;
    Move (FData [FPos + 1], Buf, Result);
    Inc (FPos, Result);
  End;

Function TStringStream.Read (const Count : Integer) : String;
var C, L : Integer;
  Begin
    L := Length (FData);
    if FPos + Count > L then
      C := L - FPos else
      C := Count;
    Result := Copy (FData, FPos + 1, C);
    Inc (FPos, C);
  End;

Procedure TStringStream.Write (const Data : TOctet);
  Begin
    if FPos = Length (FData) then
      StrAppend (FData, Char (Data)) else
      FData [FPos] := Char (Data);
    Inc (FPos);
  End;

Function TStringStream.Write (const Buf; Count : Integer) : LongInt;
  Begin
    if FPos + Count > Length (FData) then
      SetLength (FData, FPos + Count);
    Move (Buf, FData [FPos + 1], Count);
    Inc (FPos, Count);
    Result := Count;
  End;

Procedure TStringStream.Write (const Data : String);
var L : Integer;
  Begin
    L := Length (Data);
    if L = 0 then
      exit;
    if FPos + L > Length (FData) then
      SetLength (FData, FPos + L);
    Move (Data [1], FData [FPos + 1], L);
    Inc (FPos, L);
  End;

Function TStringStream.EOF : Boolean;
  Begin
    Result := FPos >= Length (FData);
  End;

Procedure TStringStream.Skip;
  Begin
    if EOF then
      raise EStream.Create ('Skip past EOF');
    Inc (FPos);
  End;

Procedure TStringStream.Skip (const Count : Integer);
  Begin
    if FPos + Count > Length (FData) then
      FPos := Length (FData) else
      Inc (FPos, Count);
  End;

Function TStringStream.Peek : TOctet;
  Begin
    if EOF then
      raise EStream.Create ('Peek past EOF');
    Result := TOctet (FData [FPos + 1]);
  End;

Function TStringStream.Peek (const Count : Integer) : String;
  Begin
    Result := Copy (FData, FPos + 1, Count);
  End;

Function TStringStream.GetPosition : Int64;
  Begin
    Result := FPos;
  End;

Procedure TStringStream.SetPosition (const NewPosition : Int64);
  Begin
    if NewPosition > Length (FData) then
      raise EStream.Create ('Seek past EOF');
    FPos := NewPosition;
  End;

Procedure TStringStream.Append;
  Begin
    FPos := Length (FData);
  End;

Function TStringStream.Remaining : Int64;
  Begin
    Result := Length (FData) - FPos;
  End;

Function TStringStream.GetStreamSize : Int64;
  Begin
    Result := Length (FData);
  End;

Procedure TStringStream.SetStreamSize (const NewSize : Int64);
  Begin
    SetLength (FData, NewSize);
  End;

Procedure TStringStream.Truncate;
  Begin
    SetLength (FData, FPos);
  End;

Function TStringStream.Locate (const S : String) : Int64;
var I : Integer;
  Begin
    I := PosNext (S, FData, FPos);
    if I = 0 then
      Result := -1 else
      Result := I - FPos - 1;
  End;

Function TStringStream.Locate (const Ch : CharSet) : Int64;
var I : Integer;
  Begin
    I := PosNext (Ch, FData, FPos);
    if I = 0 then
      Result := -1 else
      Result := I - FPos;
  End;

Function TStringStream.Match (const S : String) : Boolean;
  Begin
    Result := cStrings.Match (S, FData, FPos + 1);
  End;

{                                                                              }
{ TSocketStream                                                                }
{                                                                              }
Constructor TSocketStream.CreateEx (const Socket : {$IFDEF ICS}TSocket{$ELSE}TCustomWinSocket{$ENDIF}; const TimeOut : Integer; const CloseSocketOnDestroy : Boolean);
  Begin
    inherited Create;
    FTimeOut := TimeOut;
    FBytesRead := 0;
    FPeekBuf := '';
    FPeekBufPos := 0;
    FCloseSocket := CloseSocketOnDestroy;
    {$IFDEF ICS}
    FSocket := TWSocket.Create (nil);
    FSocket.Proto := 'tcp';
    FSocket.MultiThreaded := True;
    FSocket.HSocket := Socket;
    {$ELSE}
    FSocket := Socket;
    FSockStream := TWinSocketStream.Create (FSocket, TimeOut);
    {$ENDIF}
  End;

Constructor TSocketStream.Create (const Host : String; const Port : Integer; const TimeOut : Integer; const CloseSocketOnDestroy : Boolean);
  Begin
    inherited Create;
    FTimeOut := TimeOut;
    FBytesRead := 0;
    FPeekBuf := '';
    FPeekBufPos := 0;
    FCloseSocket := CloseSocketOnDestroy;
    {$IFDEF ICS}
    FSocket := TWSocket.Create (nil);
    FSocket.Proto := 'tcp';
    FSocket.MultiThreaded := True;
    FSocket.Port := IntToStr (Port);
    FSocket.Addr := Host;
    FSocket.Connect;
    While FSocket.State = wsConnecting do
      FSocket.ProcessMessages;
    if FSocket.State <> wsConnected then
      raise ESocketStream.Create ('Connect failed.');
    {$ELSE}
    FClient := TClientSocket.Create (nil);
    FClient.Host := Host;
    FClient.Port := Port;
    FClient.ClientType := ctBlocking;
    FClient.Open;
    FSocket := FClient.Socket;
    FSockStream := TWinSocketStream.Create (FSocket, TimeOut);
    {$ENDIF}
  End;

Destructor TSocketStream.Destroy;
  Begin
    if FCloseSocket then
      try
        CloseSocket;
      except end;
    {$IFDEF ICS}
    FreeAndNil (FSocket);
    {$ELSE}
    FreeAndNil (FClient);
    FreeAndNil (FSockStream);
    {$ENDIF}
    inherited Destroy;
  End;

Function TSocketStream.SizeUnknown : Boolean;
  Begin
    Result := True;
  End;

Procedure TSocketStream.CloseSocket;
  Begin
    if Assigned (FSocket) then
      FSocket.Close;
  End;

Function TSocketStream.EOF : Boolean;
  Begin
    {$IFDEF ICS}
    Result := FSocket.State <> wsConnected;
    {$ELSE}
    Result := not FSocket.Connected;
    {$ENDIF}
  End;

procedure TSocketStream.Append;
  Begin
  End;

Function TSocketStream.GetLocalHost : String;
  Begin
    {$IFDEF ICS}
    {$ELSE}
    Result := FSocket.LocalHost;
    {$ENDIF}
  End;

Function TSocketStream.PeekBufLeft : Integer;
  Begin
    Result := Length (FPeekBuf) - FPeekBufPos;
  End;

Function TSocketStream.FillPeekBuf (const Len : Integer) : Boolean;
const BufferClearSize = 16384;
var L : Integer;
    T : Int64;
    S : String;
  Begin
    // Clear peek buffer
    L := Length (FPeekBuf);
    if (L > 0) and (FPeekBufPos >= L) then
      begin
        FPeekBuf := '';
        FPeekBufPos := 0;
      end else
    if FPeekBufPos >= BufferClearSize then
      begin
        Move (FPeekBuf [FPeekBufPos + 1], FPeekBuf [1], L - FPeekBufPos);
        SetLength (FPeekBuf, L - FPeekBufPos);
        FPeekBufPos := 0;
      end;

    {$IFDEF ICS}
    L := Len - PeekBufLeft;

    While L > 0 do
      begin
        L := Min (L, FSocket.RcvdCount);
        if L > 0 then
          begin
            SetLength (S, L);
            SetLength (S, FSocket.Receive (@S [1], L));
            FPeekBuf := FPeekBuf + S;
          end;

        L := Len - PeekBufLeft;
        if L <= 0 then
          begin
            Result := True;
            break;
          end;

        T := StartTimer;
        Repeat
          FSocket.ProcessMessages;
          if (FSocket.State <> wsConnected) or (MillisecondsElapsed (T) >= FTimeOut) then
            begin
              Result := False;
              exit;
            end;
        Until FSocket.RcvdCount > 0;
      end;
    Result := False;
    {$ELSE}
    Repeat
      L := Len - PeekBufLeft;
      if L <= 0 then
        begin
          Result := True;
          exit;
        end;

      L := FSocket.ReceiveLength;
      if L = 0 then
        if FSockStream.WaitForData (FTimeOut) then
          L := FSocket.ReceiveLength;
      if L > 0 then
       begin
         SetLength (S, L);
         SetLength (S, FSockStream.Read (S [1], L));
         FPeekBuf := FPeekBuf + S;
         L := Length (S);
       end;
    Until L = 0;
    Result := False;
    {$ENDIF}
  End;

Function TSocketStream.Peek : TOctet;
  Begin
    if not FillPeekBuf (Sizeof (Result)) then
      raise ESocketStream.Create ('Stream read error.');
    Result := TOctet (FPeekBuf [FPeekBufPos + 1]);
  End;

Function TSocketStream.Peek (const Count : Integer) : String;
  Begin
    FillPeekBuf (Count);
    Result := System.Copy (FPeekBuf, FPeekBufPos + 1, Count);
  End;

Function TSocketStream.Match (const S : String) : Boolean;
  Begin
    FillPeekBuf (Length (S));
    Result := cStrings.Match (S, FPeekBuf, FPeekBufPos + 1);
  End;

Function TSocketStream.Locate (const S : String) : Int64;
const MaxLocateBufSize = 64 * 1024;
var I, J : Integer;
  Begin
    I := cStrings.PosNext (S, FPeekBuf, FPeekBufPos);
    if I > 0 then
      begin
        Result := I - FPeekBufPos - 1;
        exit;
      end;
    While not EOF and (Length (FPeekBuf) < MaxLocateBufSize) do
      begin
        I := PeekBufLeft;
        if not FillPeekBuf (I + 1) then
          break;
        I := Max (0, Length (FPeekBuf) - (PeekBufLeft - I) - Length (S) + 1);
        J := cStrings.PosNext (S, FPeekBuf, I);
        if J > 0 then
          begin
            Result := J - FPeekBufPos - 1;
            exit;
          end;
      end;
    Result := -1;
  End;

Function TSocketStream.Locate (const Ch : CharSet) : Int64;
const MaxLocateBufSize = 64 * 1024;
var I, J : Integer;
  Begin
    I := cStrings.PosNext (Ch, FPeekBuf, FPeekBufPos);
    if I > 0 then
      begin
        Result := I - FPeekBufPos - 1;
        exit;
      end;
    While not EOF and (Length (FPeekBuf) < MaxLocateBufSize) do
      begin
        I := PeekBufLeft;
        if not FillPeekBuf (I + 1) then
          break;
        I := Length (FPeekBuf) - (PeekBufLeft - I);
        J := cStrings.PosNext (Ch, FPeekBuf, I);
        if J > 0 then
          begin
            Result := J - FPeekBufPos - 1;
            exit;
          end;
      end;
    Result := -1;
  End;

Function TSocketStream.Read (var Buf; Count : Integer) : Integer;
type ByteBuffer = Array [0..2140000000] of Byte;
var Buffer : ByteBuffer absolute Buf;
    I, L : Integer;
  Begin
    {$IFDEF ICS}
    FillPeekBuf (Count);
    {$ENDIF}
    L := PeekBufLeft;
    if L > 0 then
      begin
        I := Min (Count, L);
        Move (FPeekBuf [FPeekBufPos + 1], Buffer [0], I);
        Inc (FPeekBufPos, I);
        Dec (Count, I);
      end else
      I := 0;
    {$IFNDEF ICS}
    if Count = 0 then
      Result := I else
      Result := I + FSockStream.Read (Buffer [I], Count);
    {$ENDIF}
  End;

Function TSocketStream.Read (const Count : Integer) : String;
var L : Integer;
  Begin
    FillPeekBuf (Count);
    Result := System.Copy (FPeekBuf, FPeekBufPos + 1, Count);
    L := Length (Result);
    Inc (FPeekBufPos, L);
    Inc (FBytesRead, L);
  End;

Function TSocketStream.Read : TOctet;
  Begin
    if not FillPeekBuf (Sizeof (Result)) then
      raise ESocketStream.Create ('Time out reading from socket.');

    Inc (FPeekBufPos);
    Result := TOctet (FPeekBuf [FPeekBufPos]);
    Inc (FBytesRead);
  End;

Procedure TSocketStream.Skip;
  Begin
    if not FillPeekBuf (1) then
      raise ESocketStream.Create ('Time out reading from socket.');
    Inc (FPeekBufPos);
    Inc (FBytesRead);
  End;

Procedure TSocketStream.Skip (const Count : Integer);
  Begin
    if not FillPeekBuf (Count) then
      raise ESocketStream.Create ('Time out reading from socket.');
    Inc (FPeekBufPos, Count);
    Inc (FBytesRead, Count);
  End;

Function TSocketStream.ReadAvailable : String;
var L : Integer;
  Begin
    {$IFDEF ICS}
    FillPeekBuf (Length (FPeekBuf) + FSocket.RcvdCount);
    Result := FPeekBuf;
    {$ELSE}
    L := FSocket.ReceiveLength;
    if L > 0 then
      begin
        SetLength (Result, L);
        SetLength (Result, FSockStream.Read (Result [1], L));
      end else
      Result := '';
    Result := cStrings.CopyFrom (FPeekBuf, FPeekBufPos + 1) + Result;
    {$ENDIF}
    FPeekBuf := '';
    FPeekBufPos := 0;
  End;

Function TSocketStream.GetPosition : Int64;
  Begin
    Result := FBytesRead;
  End;

Procedure TSocketStream.SetPosition (const NewPosition : Int64);
var P : Int64;
  Begin
    P := Position;
    if NewPosition >= P then
      Read (NewPosition - P) else
      raise EStream.Create ('TSocketStream does not support seeking backwards.');
  End;

Function TSocketStream.Write (const Buf; Count : Integer) : LongInt;
const SendBlockSize = 2048;
type ByteBuffer = Array [1..2140000000] of Byte;
var I, L   : Integer;
    Buffer : ByteBuffer absolute Buf;
  Begin
    if Count <= 0 then
      begin
        Result := 0;
        exit;
      end;
    I := 1;
    {$IFDEF ICS}
    While (I <= Count) and (FSocket.State = wsConnected) do
      begin
        L := Min (Count - I + 1, SendBlockSize);
        L := FSocket.Send (@Buffer [I], L);
        if L = 0 then
          break;
        Inc (I, L);
      end;
    {$ELSE}
    While (I <= Count) and FSocket.Connected do
      begin
        L := Min (Count - I + 1, SendBlockSize);
        L := FSockStream.Write (Buffer [I], L);
        if L = 0 then
          break;
        Inc (I, L);
      end;
    {$ENDIF}
    Result := I - 1;
  End;

Procedure TSocketStream.Write (const Data : String);
var I, L : Integer;
  Begin
    L := Length (Data);
    if L = 0 then
      exit;
    I := Write (Data [1], L);
    if I < L then
      raise EWriteError.Create ('Could not write to socket');
  End;

Procedure TSocketStream.Write (const Data : TOctet);
  Begin
    Write (Data, 1);
  End;



{                                                                              }
{ TStreamer                                                                    }
{                                                                              }
Constructor TStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create;
    Init (Stream);
  End;

Procedure TStreamer.Init (const Stream : TExStream);
  Begin
    FStream := Stream;
  End;

Destructor TStreamer.Destroy;
  Begin
    FreeAndNil (FStream);
    inherited Destroy;
  End;

Procedure TStreamer.Reset;
  Begin
    FStream.Reset;
  End;

Procedure TStreamer.ReleaseStream;
  Begin
    FStream := nil;
  End;



{                                                                              }
{ TBinaryStreamer                                                              }
{                                                                              }
{   Uses internal Delphi formats:                                              }
{     Integer       4 byte integer (LSB first)                                 }
{     Int64         8 byte integer                                             }
{     Single        4 byte FPU Single                                          }
{     Double        8 byte FPU Double                                          }
{     Extended      10 byte FPU Extended                                       }
{     Currency      8 byte FPU fixed point                                     }
{     Boolean       1 byte, 0 = False                                          }
{     DateTime      8 byte FPU Double                                          }
{     String        Integer length + String                                    }
{     ShortString   Byte length + String                                       }
{                                                                              }
Function ReadPackedShortInt (const Stream : TExStream) : ShortInt;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedWord (const Stream : TExStream) : Word;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedSmallInt (const Stream : TExStream) : SmallInt;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedLongWord (const Stream : TExStream) : LongWord;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedInteger (const Stream : TExStream) : Integer;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedInt64 (const Stream : TExStream) : Int64;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedSingle (const Stream : TExStream) : Single;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedDouble (const Stream : TExStream) : Double;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedExtended (const Stream : TExStream) : Extended;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedCurrency (const Stream : TExStream) : Currency;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedBoolean (const Stream : TExStream) : Boolean;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedDateTime (const Stream : TExStream) : TDateTime;
  Begin
    Stream.Read (Result, Sizeof (Result));
  End;

Function ReadPackedString (const Stream : TExStream) : String;
var L : Integer;
  Begin
    L := ReadPackedInteger (Stream);
    SetLength (Result, L);
    if L > 0 then
      Stream.Read (Result [1], L);
  End;

Function ReadPackedShortString (const Stream : TExStream) : ShortString;
var L : Byte;
  Begin
    L := Stream.Read;
    SetLength (Result, L);
    if L > 0 then
      Stream.Read (Result [1], L);
  End;

Procedure WritePackedShortInt (const Stream : TExStream; const D : ShortInt);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedWord (const Stream : TExStream; const D : Word);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedSmallInt (const Stream : TExStream; const D : SmallInt);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedLongWord (const Stream : TExStream; const D : LongWord);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedInteger (const Stream : TExStream; const D : Integer);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedInt64 (const Stream : TExStream; const D : Int64);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedSingle (const Stream : TExStream; const D : Single);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedDouble (const Stream : TExStream; const D : Double);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedExtended (const Stream : TExStream; const D : Extended);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedCurrency (const Stream : TExStream; const D : Currency);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedBoolean (const Stream : TExStream; const D : Boolean);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedDateTime (const Stream : TExStream; const D : TDateTime);
  Begin
    Stream.Write (D, Sizeof (D));
  End;

Procedure WritePackedString (const Stream : TExStream; const D : String);
var L : Integer;
  Begin
    L := Length (D);
    WritePackedInteger (Stream, L);
    if L > 0 then
      Stream.Write (D [1], L);
  End;

Procedure WritePackedShortString (const Stream : TExStream; const D : ShortString);
  Begin
    Stream.Write (D [0], Length (D) + 1);
  End;



{                                                                              }
{ Text parsing                                                                 }
{                                                                              }
Function SkipChar (const Stream : TExStream; const Ch : CharSet) : Boolean;
  Begin
    Result := not Stream.EOF and (Char (Stream.Peek) in Ch);
    if Result then
      Stream.Skip;
  End;

Function SkipChars (const Stream : TExStream; const Ch : CharSet) : Boolean;
  Begin
    Result := SkipChar (Stream, Ch);
    if Result then
      Repeat
      Until not SkipChar (Stream, Ch);
  End;

Function SkipWhiteSpace (const Stream : TExStream) : Boolean;
  Begin
    Result := SkipChars (Stream, cs_WhiteSpace)
  End;

Function MatchChar (const Stream : TExStream; const Ch : CharSet) : Boolean;
  Begin
    Result := not Stream.EOF and (Char (Stream.Peek) in Ch);
  End;

Function MatchWhiteSpace (const Stream : TExStream) : Boolean;
  Begin
    Result := MatchChar (Stream, cs_WhiteSpace);
  End;

Function Match (const Stream : TExStream; const Ch : Char) : Boolean;
  Begin
    Result := not Stream.EOF and (Char (Stream.Peek) = Ch);
  End;

Function ExtractChar (const Stream : TExStream; const Ch : CharSet) : String;
  Begin
    if MatchChar (Stream, Ch) then
      Result := Extract (Stream) else
      Result := '';
  End;

Function Extract (const Stream : TExStream) : Char;
  Begin
    Result := Char (Stream.Read);
  End;

Function ExtractWhiteSpace (const Stream : TExStream) : String;
  Begin
    Result := ExtractChars (Stream, cs_WhiteSpace);
  End;

Function ExtractChars (const Stream : TExStream; const Ch : CharSet) : String;
  Begin
    Result := '';
    While not Stream.EOF and (Char (Stream.Peek) in Ch) do
      StrAppend (Result, Char (Stream.Read));
  End;

Function SkipTo (const Stream : TExStream; const Delimiter : String) : Boolean;
var I : Int64;
  Begin
    I := Stream.Locate (Delimiter);
    Result := I >= 0;
    if I > 0 then
      Stream.Skip (I);
  End;

Function ExtractTo (const Stream : TExStream; const Delimiter : CharSet; const SkipDelimiter : Boolean = False) : String;
  Begin
    Result := '';
    While not Stream.EOF and not (Char (Stream.Peek) in Delimiter) do
      StrAppend (Result, Char (Stream.Read));
    if SkipDelimiter then
      Stream.Skip;
  End;

Function ExtractTo (const Stream : TExStream; const Delimiter : String; const SkipDelimiter : Boolean = False) : String;
var R : Boolean;
  Begin
    Result := '';
    R := False;
    While not Stream.EOF and not R do
      begin
        R := Stream.Peek (Length (Delimiter)) = Delimiter;
        if not R then
          StrAppend (Result, Char (Stream.Read));
      end;
    if SkipDelimiter and R then
      Stream.Skip (Length (Delimiter));
  End;

Function ExtractQuoted (const Stream : TExStream; const Quote : Char; const AllowQuotedQuote : Boolean) : String;
var Fin : Boolean;
  Begin
    if Match (Stream, Quote) then
      begin
        Extract (Stream);
        Result := '';
        Repeat
          Result := Result + ExtractTo (Stream, [Quote]);
          if Stream.EOF then
            raise EStream.Create ('No closing quote (' + Quote + ' expected)');
          Stream.Skip;
          Fin := not AllowQuotedQuote or not Match (Stream, Quote);
          if not Fin then
            begin
              Result := Result + Quote;
              Stream.Skip;
            end;
        Until Fin;
      end else
      Result := '';
  End;

Function MatchEOL (const Stream : TExStream; const EOLType : TEOLType) : Boolean;
var CR, LF : Boolean;
    Ch     : Char;
    S      : String;
  Begin
    Result := (eol_EOF in EOLType) and Stream.EOF;
    if Result then
      exit;

    try
      CR := eol_CR in EOLType;
      LF := eol_LF in EOLType;
      if CR or LF then
        begin
          Ch := Char (Stream.Peek);
          Result := CR and (Ch = ASCII_CR) or
                    LF and (Ch = ASCII_LF);
          if Result then
            exit;
        end;

      CR := eol_CRLF in EOLType;
      LF := eol_LFCR in EOLType;
      if CR or LF then
        begin
          Result := CR and Stream.Match (ASCII_CR + ASCII_LF) or
                 LF and Stream.Match (ASCII_LF + ASCII_CR);
        end;
    except
      if eol_EOF in EOLType then
        begin
          Result := True;
          exit;
        end else
        raise;
    end;
  End;

Function SkipEOL (const Stream : TExStream; const EOLType : TEOLType) : Boolean;
var CR, LF : Boolean;
    Ch     : Char;
    S      : String;
  Begin
    Result := (eol_EOF in EOLType) and Stream.EOF;
    if Result then
      exit;

    try
      CR := eol_CR in EOLType;
      LF := eol_LF in EOLType;
      if CR or LF then
        begin
          Ch := Char (Stream.Peek);
          Result := CR and (Ch = ASCII_CR) or
                    LF and (Ch = ASCII_LF);
          if Result then
            begin
              Stream.Skip;
              exit;
            end;
        end;

      CR := eol_CRLF in EOLType;
      LF := eol_LFCR in EOLType;
      if CR or LF then
        begin
          Result := CR and Stream.Match (ASCII_CR + ASCII_LF) or
                 LF and Stream.Match (ASCII_LF + ASCII_CR);
          if Result then
            Stream.Skip (2);
        end;
    except
      if eol_EOF in EOLType then
        begin
          Result := True;
          exit;
        end else
        raise;
    end;
  End;


Function ExtractEOL (const Stream : TExStream; const EOLType : TEOLType) : String;
var CR, LF : Boolean;
    Ch     : Char;
    S      : String;
  Begin
    Result := '';
    if (eol_EOF in EOLType) and Stream.EOF then
      exit;

    CR := eol_CR in EOLType;
    LF := eol_LF in EOLType;
    if CR or LF then
      begin
        Ch := Char (Stream.Peek);
        if CR and (Ch = ASCII_CR) or
           LF and (Ch = ASCII_LF) then
          begin
            Result := Extract (Stream);
            exit;
          end;
      end;

    CR := eol_CRLF in EOLType;
    LF := eol_LFCR in EOLType;
    if CR or LF then
      begin
        S := Stream.Peek (2);
        if CR and (S = ASCII_CR + ASCII_LF) or
           LF and (S = ASCII_LF + ASCII_CR) then
          begin
            Result := S;
            Stream.Skip (2);
          end;
      end;
  End;

Function LocateEOL (const Stream : TExStream; const EOLType : TEOLType; var EOLEncountered : TEOL) : Int64;

  Procedure CheckMatch (const Idx : Integer; const EOL : TEOL);
    Begin
      if (Idx >= 0) and ((Result = -1) or (Idx < Result)) then
        begin
          EOLEncountered := EOL;
          Result := Idx;
        end;
    End;

  Begin
    Result := -1;
    if eol_CR in EOLType then
      CheckMatch (Stream.Locate ([ASCII_CR]), eol_CR);
    if eol_LF in EOLType then
      CheckMatch (Stream.Locate ([ASCII_LF]), eol_LF);
    if eol_CRLF in EOLType then
      CheckMatch (Stream.Locate (ASCII_CR + ASCII_LF), eol_CRLF);
    if eol_LFCR in EOLType then
      CheckMatch (Stream.Locate (ASCII_LF + ASCII_CR), eol_LFCR);
    if (Result = -1) and (eol_EOF in EOLType) then
      begin
        if Stream.EOF then
          Result := 0 else
          if not Stream.SizeUnknown then
            Result := Stream.Remaining else
            Result := -1;
        EOLEncountered := eol_EOF;
      end;
  End;

Function ExtractLine (const Stream : TExStream; const EOLType : TEOLType; const MaxLineLength : Integer) : String;
var I : Int64;
    EOL : TEOL;
  Begin
    I := LocateEOL (Stream, EOLType, EOL);
    if I >= 0 then
      begin
        Result := Stream.Read (I);
        Case EOL of
          eol_CRLF, eol_LFCR : Stream.Skip (2);
          eol_CR, eol_LF : Stream.Skip;
        end;
      end else
      raise EReadError.Create ('Stream read error.');
  End;

Function SkipLine (const Stream : TExStream; const EOLType : TEOLType) : Boolean;
  Begin
    While not MatchEOL (Stream, EOLType) do
      Stream.Skip;
    Result := SkipEOL (Stream, EOLType);
  End;

Function MatchLine (const Stream : TExStream; const S : String; const EOLType : TEOLType) : Boolean;
var T : String;
    P : Int64;
  Begin
    P := Stream.Position;
    T := Stream.Extract (S);
    Result := (T = S) and MatchEOL (Stream);
    if Result then
      SkipEOL (Stream, EOLType) else
      Stream.Position := P;
  End;

{ Multi-line functions                                                         }
Procedure StreamEmptyLineTerminated (const Source, Dest : TExStream; const EOLType : TEOLType);
var S : String;
  Begin
    While not Source.EOF do
      begin
        S := ExtractLine (Source, EOLType);
        if (S = '') or (S [1] = '.') then
          S := '.' + S;
        Dest.Write (S + c_CRLF);
      end;
    Dest.Write (c_CRLF);
  End;

Procedure StreamEmptyLineTerminated (const Source : String; const Dest : TExStream);
var S : StringArray;
    I : Integer;
  Begin
    if Match (c_CRLF, Source, Length (Source) - 1) then
      S := Split (CopyLeft (Source, Length (Source) - 2), c_CRLF) else
      S := Split (Source, c_CRLF);
    For I := 0 to Length (S) - 1 do
      begin
        if (S [I] = '') or (S [I][1] = '.') then
          S [I] := '.' + S [I];
        Dest.Write (S [I] + c_CRLF);
      end;
    Dest.Write (c_CRLF);
  End;

Function EncodeEmptyLineTerminated (const S : String) : String;
var T : StringArray;
    I : Integer;
  Begin
    if Match (c_CRLF, S, Length (S) - 1) then
      T := Split (CopyLeft (S, Length (S) - 2), c_CRLF) else
      T := Split (S, c_CRLF);
    For I := 0 to Length (T) - 1 do
      if (T [I] = '') or (T [I][1] = '.') then
        Result := Result + '.' + T [I] + c_CRLF else
        Result := Result + T [I] + c_CRLF;
    Result := Result + c_CRLF;
  End;

Function ExtractEmptyLineTerminated (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : String;
var R : Boolean;
    S : String;
  Begin
    Result := '';
    R := False;
    While not R and not Stream.EOF do
      begin
        S := ExtractLine (Stream, EOLType);
        R := S = '';
        if not R then
          if S [1] = '.' then
            Result := Result + CopyFrom (S, 2) + c_CRLF else
            Result := Result + S + c_CRLF;
      end;
  End;

Procedure StreamDotLineTerminated (const Source, Dest : TExStream; const EOLType : TEOLType; const ProgressCallback : TStreamProgressCallback);
var S : String;
    C : Int64;
    A : Boolean;
  Begin
    C := Source.Position;
    A := False;
    While not Source.EOF and not A do
      begin
        S := ExtractLine (Source, EOLType);
        if (Length (S) > 0) and (S [1] = '.') then
          S := '.' + S;
        Dest.Write (S + c_CRLF);
        if Assigned (ProgressCallback) then
          ProgressCallBack (Source.Position - C, A);
      end;
    if A then
      exit;
    Dest.Write ('.' + c_CRLF);
  End;

Procedure StreamDotLineTerminated (const Source : String; const Dest : TExStream; const ProgressCallback : TStreamProgressCallback);
var S : StringArray;
    I : Integer;
    A : Boolean;
    P : Int64;
  Begin
    if Match (c_CRLF, Source, Length (Source) - 1) then
      S := Split (CopyLeft (Source, Length (Source) - 2), c_CRLF) else
      S := Split (Source, c_CRLF);
    A := False;
    P := 0;
    For I := 0 to Length (S) - 1 do
      begin
        P := P + Length (S [I]) + Length (c_CRLF);
        if (Length (S [I]) > 0) and (S [I][1] = '.') then
          S [I] := '.' + S [I];
        Dest.Write (S [I] + c_CRLF);
        if Assigned (ProgressCallback) then
          begin
            ProgressCallBack (P, A);
            if A then
              exit;
          end;
      end;
    Dest.Write ('.' + c_CRLF);
  End;

Function EncodeDotLineTerminated (const S : String) : String;
var T : StringArray;
    I : Integer;
  Begin
    if Match (c_CRLF, S, Length (S) - 1) then
      T := Split (CopyLeft (S, Length (S) - 2), c_CRLF) else
      T := Split (S, c_CRLF);
    For I := 0 to Length (T) - 1 do
      if (Length (T [I]) > 0) and (T [I][1] = '.') then
        Result := Result + '.' + T [I] + c_CRLF else
        Result := Result + T [I] + c_CRLF;
    Result := Result + '.' + c_CRLF;
  End;

Function ExtractDotLineTerminated (const Stream : TExStream; const EOLType : TEOLType = [eol_EOF, eol_CRLF]) : String;
var R : Boolean;
    S : String;
  Begin
    Result := '';
    R := False;
    While not R and not Stream.EOF do
      begin
        S := ExtractLine (Stream, EOLType);
        R := S = '.';
        if not R then
          begin
            if (Length (S) > 0) and (S [1] = '.') then
              System.Delete (S, 1, 1);
            S := S + c_CRLF;
            Result := Result + S;
          end;
      end;
  End;

Function ExtractMultiLine (const Stream : TExStream; const EOLType : TEOLType) : StringArray;
  Begin
    While not Stream.EOF do
      Append (Result, ExtractLine (Stream, EOLType));
  End;

Function ExtractToEOF (const Stream : TExStream) : String;
  Begin
    Result := '';
    While not Stream.EOF do
      Result := Result + Stream.Read (Max (1, Stream.Remaining));
  End;

Function MatchKeyword (const Stream : TExStream; const Keywords : StringArray; const RightDelimiter : CharSet) : Integer;
var I, L : Integer;
    S    : String;
  Begin
    For I := 0 to High (Keywords) do
      if Stream.Match (Keywords [I]) then
        begin
          S := Stream.Peek (Length (Keywords [I]) + 1);
          L := Length (S);
          if (L = Length (Keywords [I])) or (RightDelimiter = []) or (S [L] in RightDelimiter) then
            begin
              Result := I;
              exit;
            end;
        end;
    Result := -1;
  End;

Function ExtractKeyword (const Stream : TExStream; const Keywords : StringArray; const RightDelimiter : CharSet) : String;
var I : Integer;
  Begin
    I := MatchKeyword (Stream, Keywords, RightDelimiter);
    if I = -1 then
      Result := '' else
      Result := Stream.Read (Length (Keywords [I]));
  End;



{                                                                              }
{ TTextParser                                                                  }
{                                                                              }
Constructor TTextParser.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream);
    FEOLType := [eol_CRLF, eol_EOF];
    FLineNr := 1;
    FLastNewLine := 0;
    FMaxLineLength := -1;
    FWhiteSpace := cs_WhiteSpace;
    FCaseSensitive := False;
    FLocaleSensitive := False;
  End;

Function TTextParser.EOF : Boolean;
  Begin
    Result := FStream.EOF;
  End;

Function TTextParser.GetCol : Integer;
  Begin
    Result := FStream.GetPosition - FLastNewLine + 1;
  End;

{ Match                                                                        }
Function TTextParser.Match (const Txt : String) : Boolean;
var L, R : Integer;
    S    : String;
  Begin
    L := Length (Txt);
    R := FStream.Remaining;
    if (R >= 0) and (R < L) then
      begin
        Result := False;
        exit;
      end;

    S := FStream.Peek (L);
    if FCaseSensitive and FLocaleSensitive then
      Result := CaseLocaleInsensitiveMatch (Txt, S) else
    if FCaseSensitive then
      Result := CaseInsensitiveMatch (Txt, S) else
    if FLocaleSensitive then
      Result := LocaleInsensitiveMatch (Txt, S) else
      Result := Txt = S;
  End;

Function TTextParser.Extract (const Txt : String) : String;
  Begin
    if Match (Txt) then
      Result := FStream.Read (Length (Txt)) else
      Result := '';
  End;

Function TTextParser.MatchLine (const Txt : String) : Boolean;
var T : String;
    P : Int64;
  Begin
    P := FStream.Position;
    T := Extract (Txt);
    Result := (T <> '') and MatchEOL (FStream, FEOLType);
    if Result then
      SkipEOL (FStream, FEOLType) else
      FStream.Position := P;
  End;

Function TTextParser.GetBookmark : Int64;
  Begin
    Result := FStream.Position;
  End;

Procedure TTextParser.GotoBookmark (const B : Int64);
  Begin
    FStream.Position := B;
  End;

Procedure TTextParser.Writeln (const S : String);
  Begin
    FStream.Write (S + c_CRLF);
  End;

Function TTextParser.Readln : String;
  Begin
    Result := ExtractLine (FStream, FEOLType, FMaxLineLength);
  End;


{                                                                              }
{ TRecordStreamer                                                              }
{                                                                              }
Constructor TRecordStreamer.Create (const Stream : TExStream; const RecordSize : Integer; const FirstRecordOffset : Int64);
  Begin
    if Stream.SequentialOnly then
      raise EStream.Create ('RecordStreamer must have a random access stream.');
    inherited Create (Stream);
    FRecordSize := RecordSize;
    FRecordOffset := FirstRecordOffset;
  End;

Function TRecordStreamer.GetRecordCount : Int64;
  Begin
    Result := (FStream.Size - FRecordOffset) div FRecordSize;
  End;

Procedure TRecordStreamer.SetRecordCount (const NewRecordCount : Int64);
  Begin
    FStream.SetStreamSize (FRecordOffset + NewRecordCount * FRecordSize);
  End;

Procedure TRecordStreamer.SetRecordSize (const RecordSize : Integer);
  Begin
    if RecordSize <= 0 then
      raise EStream.Create ('Invalid record size') else
      FRecordSize := RecordSize;
  End;

Procedure TRecordStreamer.Seek (const Idx : Int64);
  Begin
    FStream.Position := FRecordOffset + Idx * FRecordSize;
  End;

Procedure TRecordStreamer.Read (const Idx : Int64; const Count : Integer; var Buf);
  Begin
    Seek (Idx);
    FStream.Read (Buf, Count * FRecordSize);
  End;

Procedure TRecordStreamer.Write (const Idx : Int64; const Count : Integer; const Buf);
  Begin
    Seek (Idx);
    FStream.Write (Buf, Count * FRecordSize);
  End;

Procedure TRecordStreamer.Insert (const Idx : Int64; const Count : Integer);
  Begin
    Seek (Idx);
    FStream.Insert (Count * FRecordSize);
  End;

Procedure TRecordStreamer.Delete (const Idx : Int64; const Count : Integer);
  Begin
    Seek (Idx);
    FStream.Delete (Count * FRecordSize);
  End;

Procedure TRecordStreamer.Clear (const Idx : Int64; const Count : Integer);
const BufSize = 1024;
var Buffer  : String;
    L, C, I : Integer;
  Begin
    if Count = 0 then
      exit;
    C := Min (Count, BufSize);
    L := FRecordSize * C;
    SetLength (Buffer, L);
    FillChar (Buffer [1], L, #0);
    For I := 0 to Count div C do
      Write (Idx + I * C, C, Buffer [1]);
  End;

{                                                                              }
{ TIntegerArrayStreamer                                                        }
{                                                                              }
Constructor TIntegerArrayStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream, Sizeof (Integer));
  End;

Procedure TIntegerArrayStreamer.SetItem (const Idx : Integer; const Value : Integer);
  Begin
    Write (Idx, 1, Value);
  End;

Function TIntegerArrayStreamer.GetItem (const Idx : Integer) : Integer;
  Begin
    Read (Idx, 1, Result);
  End;

Procedure TIntegerArrayStreamer.Fill (const Value : Integer);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : IntegerArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    A := DupInteger (Value, C);
    For I := 0 to (RC - 1) div C do
      Write (I * C, Min (C, RC - I * C), A [0]);
  End;

Procedure TIntegerArrayStreamer.FillInc (const StartValue : Integer; const Increment : Integer);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : IntegerArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    For I := 0 to (RC - 1) div C do
      begin
        A := RangeInteger (StartValue + I * C * Increment, C, Increment);
        Write (I * C, Min (C, RC - I * C), A [0]);
      end;
  End;

Function TIntegerArrayStreamer.GetAsIntegerArray : IntegerArray;
var L : Integer;
  Begin
    L := Count;
    SetLength (Result, L);
    Read (0, L, Result [0]);
  End;

Procedure TIntegerArrayStreamer.SetAsIntegerArray (const V : IntegerArray);
var L : Integer;
  Begin
    L := Length (V);
    Count := L;
    Write (0, L, V [0]);
  End;

Function TIntegerArrayStreamer.PosNext (const Find : Integer; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Integer;
  Begin
    if IsSortedAscending then // binary search
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first
          begin
            L := 0;
            H := Count - 1;
            Repeat
              I := (L + H) div 2;
              D := Item [I];
              if D = Find then
                begin
                  While (I > 0) and (Item [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Count - 1 then
            Result := -1 else
            if Item [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Count - 1 do
          if Item [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;



{                                                                              }
{ TInt64ArrayStreamer                                                          }
{                                                                              }
Constructor TInt64ArrayStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream, Sizeof (Int64));
  End;

Procedure TInt64ArrayStreamer.SetItem (const Idx : Integer; const Value : Int64);
  Begin
    Write (Idx, 1, Value);
  End;

Function TInt64ArrayStreamer.GetItem (const Idx : Integer) : Int64;
  Begin
    Read (Idx, 1, Result);
  End;

Procedure TInt64ArrayStreamer.Fill (const Value : Int64);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : Int64Array;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    A := DupInt64 (Value, C);
    For I := 0 to (RC - 1) div C do
      Write (I * C, Min (C, RC - I * C), A [0]);
  End;

Procedure TInt64ArrayStreamer.FillInc (const StartValue : Int64; const Increment : Int64);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : Int64Array;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    For I := 0 to (RC - 1) div C do
      begin
        A := RangeInt64 (StartValue + I * C * Increment, C, Increment);
        Write (I * C, Min (C, RC - I * C), A [0]);
      end;
  End;

Function TInt64ArrayStreamer.GetAsInt64Array : Int64Array;
var L : Integer;
  Begin
    L := Count;
    SetLength (Result, L);
    Read (0, L, Result [0]);
  End;

Procedure TInt64ArrayStreamer.SetAsInt64Array (const V : Int64Array);
var L : Integer;
  Begin
    L := Length (V);
    Count := L;
    Write (0, L, V [0]);
  End;

Function TInt64ArrayStreamer.PosNext (const Find : Int64; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Int64;
  Begin
    if IsSortedAscending then // binary search
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first
          begin
            L := 0;
            H := Count - 1;
            Repeat
              I := (L + H) div 2;
              D := Item [I];
              if D = Find then
                begin
                  While (I > 0) and (Item [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Count - 1 then
            Result := -1 else
            if Item [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Count - 1 do
          if Item [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;



{                                                                              }
{ TSingleArrayStreamer                                                         }
{                                                                              }
Constructor TSingleArrayStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream, Sizeof (Single));
  End;

Procedure TSingleArrayStreamer.SetItem (const Idx : Integer; const Value : Single);
  Begin
    Write (Idx, 1, Value);
  End;

Function TSingleArrayStreamer.GetItem (const Idx : Integer) : Single;
  Begin
    Read (Idx, 1, Result);
  End;

Procedure TSingleArrayStreamer.Fill (const Value : Single);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : SingleArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    A := DupSingle (Value, C);
    For I := 0 to (RC - 1) div C do
      Write (I * C, Min (C, RC - I * C), A [0]);
  End;

Procedure TSingleArrayStreamer.FillInc (const StartValue : Single; const Increment : Single);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : SingleArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    For I := 0 to (RC - 1) div C do
      begin
        A := RangeSingle (StartValue + I * C * Increment, C, Increment);
        Write (I * C, Min (C, RC - I * C), A [0]);
      end;
  End;

Function TSingleArrayStreamer.GetAsSingleArray : SingleArray;
var L : Integer;
  Begin
    L := Count;
    SetLength (Result, L);
    Read (0, L, Result [0]);
  End;

Procedure TSingleArrayStreamer.SetAsSingleArray (const V : SingleArray);
var L : Integer;
  Begin
    L := Length (V);
    Count := L;
    Write (0, L, V [0]);
  End;

Function TSingleArrayStreamer.PosNext (const Find : Single; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Single;
  Begin
    if IsSortedAscending then // binary search
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first
          begin
            L := 0;
            H := Count - 1;
            Repeat
              I := (L + H) div 2;
              D := Item [I];
              if D = Find then
                begin
                  While (I > 0) and (Item [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Count - 1 then
            Result := -1 else
            if Item [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Count - 1 do
          if Item [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;



{                                                                              }
{ TDoubleArrayStreamer                                                         }
{                                                                              }
Constructor TDoubleArrayStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream, Sizeof (Double));
  End;

Procedure TDoubleArrayStreamer.SetItem (const Idx : Integer; const Value : Double);
  Begin
    Write (Idx, 1, Value);
  End;

Function TDoubleArrayStreamer.GetItem (const Idx : Integer) : Double;
  Begin
    Read (Idx, 1, Result);
  End;

Procedure TDoubleArrayStreamer.Fill (const Value : Double);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : DoubleArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    A := DupDouble (Value, C);
    For I := 0 to (RC - 1) div C do
      Write (I * C, Min (C, RC - I * C), A [0]);
  End;

Procedure TDoubleArrayStreamer.FillInc (const StartValue : Double; const Increment : Double);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : DoubleArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    For I := 0 to (RC - 1) div C do
      begin
        A := RangeDouble (StartValue + I * C * Increment, C, Increment);
        Write (I * C, Min (C, RC - I * C), A [0]);
      end;
  End;

Function TDoubleArrayStreamer.GetAsDoubleArray : DoubleArray;
var L : Integer;
  Begin
    L := Count;
    SetLength (Result, L);
    Read (0, L, Result [0]);
  End;

Procedure TDoubleArrayStreamer.SetAsDoubleArray (const V : DoubleArray);
var L : Integer;
  Begin
    L := Length (V);
    Count := L;
    Write (0, L, V [0]);
  End;

Function TDoubleArrayStreamer.PosNext (const Find : Double; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Double;
  Begin
    if IsSortedAscending then // binary search
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first
          begin
            L := 0;
            H := Count - 1;
            Repeat
              I := (L + H) div 2;
              D := Item [I];
              if D = Find then
                begin
                  While (I > 0) and (Item [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Count - 1 then
            Result := -1 else
            if Item [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Count - 1 do
          if Item [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;



{                                                                              }
{ TExtendedArrayStreamer                                                       }
{                                                                              }
Constructor TExtendedArrayStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream, Sizeof (Extended));
  End;

Procedure TExtendedArrayStreamer.SetItem (const Idx : Integer; const Value : Extended);
  Begin
    Write (Idx, 1, Value);
  End;

Function TExtendedArrayStreamer.GetItem (const Idx : Integer) : Extended;
  Begin
    Read (Idx, 1, Result);
  End;

Procedure TExtendedArrayStreamer.Fill (const Value : Extended);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : ExtendedArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    A := DupExtended (Value, C);
    For I := 0 to (RC - 1) div C do
      Write (I * C, Min (C, RC - I * C), A [0]);
  End;

Procedure TExtendedArrayStreamer.FillInc (const StartValue : Extended; const Increment : Extended);
const BlockSize = 1024;
var C, I : Integer;
    RC   : Int64;
    A    : ExtendedArray;
  Begin
    RC := RecordCount;
    C := Min (RC, BlockSize);
    For I := 0 to (RC - 1) div C do
      begin
        A := RangeExtended (StartValue + I * C * Increment, C, Increment);
        Write (I * C, Min (C, RC - I * C), A [0]);
      end;
  End;

Function TExtendedArrayStreamer.GetAsExtendedArray : ExtendedArray;
var L : Integer;
  Begin
    L := Count;
    SetLength (Result, L);
    Read (0, L, Result [0]);
  End;

Procedure TExtendedArrayStreamer.SetAsExtendedArray (const V : ExtendedArray);
var L : Integer;
  Begin
    L := Length (V);
    Count := L;
    Write (0, L, V [0]);
  End;

Function TExtendedArrayStreamer.PosNext (const Find : Extended; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : Extended;
  Begin
    if IsSortedAscending then // binary search
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first
          begin
            L := 0;
            H := Count - 1;
            Repeat
              I := (L + H) div 2;
              D := Item [I];
              if D = Find then
                begin
                  While (I > 0) and (Item [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Count - 1 then
            Result := -1 else
            if Item [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Count - 1 do
          if Item [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;




{                                                                              }
{ TShortStringArrayStreamer                                                    }
{                                                                              }
Constructor TShortStringArrayStreamer.Create (const Stream : TExStream; const StringLen : Byte = 255);
  Begin
    inherited Create (Stream, StringLen + 1);
  End;

Procedure TShortStringArrayStreamer.SetItem (const Idx : Integer; const Value : ShortString);
  Begin
    Write (Idx, 1, Value [0]);
  End;

Function TShortStringArrayStreamer.GetItem (const Idx : Integer) : ShortString;
  Begin
    Read (Idx, 1, Result [0]);
  End;

Procedure TShortStringArrayStreamer.Fill (const Value : ShortString);
var I : Integer;
  Begin
    For I := 0 to RecordCount - 1 do
      Write (I, 1, Value [0]);
  End;
Function TShortStringArrayStreamer.PosNext (const Find : ShortString; const PrevPos : Integer; const IsSortedAscending : Boolean) : Integer;
var I, L, H : Integer;
    D       : ShortString;
  Begin
    if IsSortedAscending then // binary search
      begin
        if Max (PrevPos + 1, 0) = 0 then // find first
          begin
            L := 0;
            H := Count - 1;
            Repeat
              I := (L + H) div 2;
              D := Item [I];
              if D = Find then
                begin
                  While (I > 0) and (Item [I - 1] = Find) do
                    Dec (I);
                  Result := I;
                  exit;
                end else
              if D > Find then
                H := I - 1 else
                L := I + 1;
            Until L > H;
            Result := -1;
          end else // find next
          if PrevPos >= Count - 1 then
            Result := -1 else
            if Item [PrevPos + 1] = Find then
              Result := PrevPos + 1 else
              Result := -1;
      end else
      begin // linear search
        For I := Max (PrevPos + 1, 0) to Count - 1 do
          if Item [I] = Find then
            begin
              Result := I;
              exit;
            end;
        Result := -1;
      end;
  End;




{                                                                              }
{ TBitArrayStreamer                                                            }
{                                                                              }
type
  TBitArrayHeader = packed record
    Count : Int64;
  end;

Function GetBitArrayHeader (const S : TExStream) : TBitArrayHeader;
  Begin
    S.Position := 0;
    S.Read (Result, Sizeof (Result));
  End;

Constructor TBitArrayStreamer.Create (const Stream : TExStream);
  Begin
    inherited Create (Stream, Sizeof (Byte), Sizeof (TBitArrayHeader));
  End;

Procedure TBitArrayStreamer.SetCount (const NewCount : Int64);
  Begin
    SetRecordCount ((NewCount + 7) div 8);
    FStream.Position := 0;
    FStream.Write (NewCount, Sizeof (Int64));
  End;

Function TBitArrayStreamer.GetCount : Int64;
  Begin
    FStream.Position := 0;
    FStream.Read (Result, Sizeof (Int64));
  End;

Procedure TBitArrayStreamer.SetItem (const Idx : Int64; const Value : Boolean);
var B, C : Byte;
    I    : Int64;
  Begin
    I := Idx div 8;
    C := 1 shl (Idx mod 8);
    Read (I, 1, B);
    if Value then
      B := B or C else
      B := B and not C;
    Write (I, 1, B);
  End;

Function TBitArrayStreamer.GetItem (const Idx : Int64) : Boolean;
var B : Byte;
  Begin
    Read (Idx div 8, 1, B);
    Result := B or (1 shl (Idx mod 8)) <> 0;
  End;

Procedure TBitArrayStreamer.FillRange (const LoIdx, HiIdx : Int64; const Value : Boolean);
const BlockSize = 1024;
var C, I  : Integer;
    RC, J : Int64;
    A     : ByteArray;
    H, L  : Int64;
  Begin
    H := Min (HiIdx, Count);
    L := Max (0, LoIdx);
    RC := H - L + 1;
    C := Min (RC, BlockSize);
    if Value then
      A := DupByte ($FF, C) else
      A := DupByte ($00, C);
    For I := 0 to (RC - 1) div C do
      begin
        J := I * C;
        Write (L + J, Min (C, RC - J), A [0]);
      end;
  End;

Procedure TBitArrayStreamer.Fill (const Value : Boolean);
  Begin
    FillRange (0, MaxInt64, Value);
  End;

Function TBitArrayStreamer.CompareRange (const LoIdx, HiIdx : Int64; const Value : Boolean) : Boolean;
  Begin
  End;

Procedure TBitArrayStreamer.Invert;
  Begin
  End;

Function TBitArrayStreamer.Find (const Value : Boolean; const Start : Integer; const FindForward : Boolean) : Integer;
  Begin
  End;

Function TBitArrayStreamer.FindRange (const Value : Boolean; const Start : Integer; const Count : Integer; const FindForward : Boolean) : Integer;
  Begin
  End;

end.

